//file.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "private.h"

static int _openvio(uniraum_file_t * state, struct roar_vio_calls * vio, int flags) {
 char buf[_RAUM_MAGIC_LEN];
 ssize_t ret;

 if ( vio == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( roar_vio_open_re(&(state->vio_re), vio) == -1 )
  return -1;

 state->vio     = &(state->vio_re);
 state->o_flags = flags;
 state->error   = ROAR_ERROR_NONE;
 state->refc    = 1;
 state->execed  = -1;

 switch (flags & (O_RDONLY|O_WRONLY|O_RDWR)) {
  case O_RDONLY:
  case O_RDWR:
    ret = roar_vio_read(vio, buf, _RAUM_MAGIC_LEN);
    if ( ret != _RAUM_MAGIC_LEN ) {
     if ( ret != -1 )
      roar_error = ROAR_ERROR_PROTO;
     return -1;
    }
    if ( memcmp(buf, _RAUM_MAGIC, _RAUM_MAGIC_LEN) != 0 ) {
     roar_error = ROAR_ERROR_BADMAGIC;
     return -1;
    }
   break;
  case O_WRONLY:
    ret = roar_vio_write(vio, _RAUM_MAGIC, _RAUM_MAGIC_LEN);
    if ( ret != _RAUM_MAGIC_LEN) {
     return -1;
    }
   break;
 }

 return 0;
}

uniraum_file_t * uniraum_openvio(struct roar_vio_calls * vio, int flags) {
 uniraum_file_t * state = roar_mm_malloc(sizeof(uniraum_file_t));
 int err;

 if ( state == NULL )
  return NULL;

 memset(state, 0, sizeof(uniraum_file_t));

 if ( _openvio(state, vio, flags) == -1 ) {
  err = roar_error;
  roar_mm_free(state);
  roar_error = err;
  return NULL;
 }

 return state;
}

uniraum_file_t * uniraum_opendstr(const char * dstr, struct roar_vio_defaults * def, int flags) {
 struct roar_vio_defaults def_store;
 uniraum_file_t * state = roar_mm_malloc(sizeof(uniraum_file_t));
 int err;

 if ( state == NULL )
  return NULL;

 memset(state, 0, sizeof(uniraum_file_t));

 if ( def == NULL ) {
  def = &def_store;
  if ( roar_vio_dstr_init_defaults(def, ROAR_VIO_DEF_TYPE_NONE, flags, 0644) == -1 ) {
   err = roar_error;
   roar_mm_free(state);
   roar_error = err;
   return NULL;
  }
 }

 if ( roar_vio_open_dstr(&(state->vio_store), (char*)dstr, def, 1) == -1 ) {
  err = roar_error;
  roar_mm_free(state);
  roar_error = err;
  return NULL;
 }

 if ( _openvio(state, &(state->vio_store), flags) == -1 ) {
  err = roar_error;
  roar_vio_close(&(state->vio_store));
  roar_mm_free(state);
  roar_error = err;
  return NULL;
 }

 return state;
}

int              uniraum_ref(uniraum_file_t * state) {
 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 } else {
  state->refc++;
  return 0;
 }
}

int              uniraum_unref(uniraum_file_t * state) {
 uniraum_req_t * req;

 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 state->refc--;

 if ( state->refc == 0 ) {
  if ( (state->o_flags & O_WRONLY) || (state->o_flags & O_RDWR) ) {
   if ( state->execed == -1 ) {
    req = uniraum_reqeof();
    uniraum_reqout(state, req);
    uniraum_requnref(req);
   }

   if ( state->req_queue_out_fill != 0 ) {
    uniraum_write(state);
    roar_vio_sync(state->vio);
    uniraum_write(state);
   }
  }

  if ( state->req_queue_out_fill != 0 ) {
   state->refc++;
   // error is set by uniraum_write().
   return -1;
  }

  roar_vio_close(state->vio);
  roar_mm_free(state);
 }

 return 0;
}

int              uniraum_destroy(uniraum_file_t * state) {
 size_t i;

 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->refc != 1 ) {
  state->error = ROAR_ERROR_BUSY;
  return -1;
 }

 for (i = 0; i < state->req_queue_out_fill; i++) {
  uniraum_requnref(state->req_queue_out[i]);
 }

 return uniraum_unref(state);
}

int              uniraum_error(uniraum_file_t * state) {
 if ( state == NULL ) {
  return roar_error;
 } else {
  return state->error;
 }
}

int              uniraum_read(uniraum_file_t * state) {
 uniraum_req_t * req;
 char * p = NULL;
 int ret;

 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->execed != -1 ) {
  state->error = ROAR_ERROR_INVAL;
  return -1;
 }

 if ( state->req_queue_in_fill == _REQ_QUEUE_LEN ) {
  state->error = ROAR_ERROR_AGAIN;
  return -1;
 }

 req = uniraum_reqnewempty();

 ret = roar_vrecv_message(state->vio, &(req->mes), &p);
 req->long_data = p;

 if ( ret == -1 ) {
  state->error = roar_error;
  uniraum_requnref(req);
  return -1;
 }

 if ( req->mes.cmd == ROAR_CMD_EXEC_STREAM ) {
  state->execed = req->mes.stream;
 }

 state->req_queue_in[state->req_queue_in_fill++] = req;

 return 0;
}

uniraum_req_t *  uniraum_reqin(uniraum_file_t * state) {
 uniraum_req_t * ret;
 size_t i;

 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return NULL;
 }

 if ( state->req_queue_in_fill == 0 ) {
  state->error = ROAR_ERROR_AGAIN;
  return NULL;
 }

 ret = state->req_queue_in[0];

 state->req_queue_in_fill--;

 for (i = 0; i < state->req_queue_in_fill; i++) {
  state->req_queue_in[i] = state->req_queue_in[i+1];
 }

 return ret;
}

int              uniraum_reqout(uniraum_file_t * state, uniraum_req_t * req) {
 int err;

 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( req == NULL ) {
  state->error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->execed != -1 ) {
  state->error = ROAR_ERROR_INVAL;
  return -1;
 }

 if ( state->req_queue_out_fill == _REQ_QUEUE_LEN ) {
  err = state->error;
  uniraum_write(state);
  state->error = err;
 }

 if ( state->req_queue_out_fill == _REQ_QUEUE_LEN ) {
  state->error = ROAR_ERROR_AGAIN;
  return -1;
 }

 if ( uniraum_reqref(req) == -1 ) {
  state->error = roar_error;
  return -1;
 }

 state->req_queue_out[state->req_queue_out_fill++] = req;

 if ( uniraum_reqget_type(req) == ROAR_CMD_EXEC_STREAM ) {
  state->execed = uniraum_reqget_stream(req);
 }

 err = state->error;
 uniraum_write(state);
 state->error = err;

 return 0;
}

int              uniraum_write(uniraum_file_t * state) {
 size_t i, j;
 int ret = 0;
 int err;

 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->req_queue_out_fill == 0 )
  return 0;

 err = roar_error;
 roar_err_clear();

 for (i = 0; ret == 0 && i < state->req_queue_out_fill; i++) {
  ret = roar_vsend_message(state->vio, &(state->req_queue_out[i]->mes), state->req_queue_out[i]->long_data);
  if ( ret == 0 )
   uniraum_requnref(state->req_queue_out[i]);
 }

 for (j = 0; j < i; j++) {
  state->req_queue_out[j] = state->req_queue_out[j+i];
 }

 state->req_queue_out_fill -= i;

 state->error = roar_error;
 roar_error = err;

 return 0;
}

int              uniraum_get_execed(uniraum_file_t * state) {
 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 return state->execed;
}

ssize_t          uniraum_write_execed(uniraum_file_t * state, const void * buf, size_t len) {
 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( buf == NULL ) {
  state->error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->execed == -1 ) {
  state->error = ROAR_ERROR_INVAL;
  return -1;
 }

 if ( state->req_queue_out_fill != 0 ) {
  uniraum_write(state);
  if ( state->req_queue_out_fill != 0 ) {
   return -1;
  }
 }

 return roar_vio_write(state->vio, (void*)buf, len);
}

ssize_t          uniraum_read_execed (uniraum_file_t * state,       void * buf, size_t len) {
 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( buf == NULL ) {
  state->error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->execed == -1 ) {
  state->error = ROAR_ERROR_INVAL;
  return -1;
 }

 return roar_vio_read(state->vio, buf, len);
}

uniraum_seekpos_t uniraum_get_seekpos(uniraum_file_t * state) {
 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->req_queue_out_fill != 0 )
  uniraum_write(state);

 if ( state->req_queue_out_fill != 0 || state->req_queue_in_fill != 0 ) {
  roar_error = ROAR_ERROR_INVAL;
  return -1;
 }

 return (uniraum_seekpos_t)roar_vio_lseek(state->vio, 0, SEEK_CUR);
}

int               uniraum_seek(uniraum_file_t * state, uniraum_seekpos_t pos, uniraum_whence_t whence) {
 off_t ret;

 if ( state == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( state->req_queue_out_fill != 0 )
  uniraum_write(state);

 if ( state->req_queue_out_fill != 0 || state->req_queue_in_fill != 0 ) {
  roar_error = ROAR_ERROR_INVAL;
  return -1;
 }

 ret = roar_vio_lseek(state->vio, pos, whence);
 if ( ret == (off_t)-1 )
  return -1;

 return 0;
}

//ll
