//mapping_pcm.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "private.h"

#ifdef HAVE_LIBOGG

#include <ogg/ogg.h>

struct ogg_mapping {
 ogg_sync_state oy;
 int oy_opened;
};

int uniraum_mapper_ogg_init(uniraum_mapping_t * mapping) {
 struct ogg_mapping * self = roar_mm_malloc(sizeof(struct ogg_mapping));

 if ( self == NULL )
  return -1;

 memset(self, 0, sizeof(struct ogg_mapping));

 mapping->state.vp = self;

 return 0;
}

int uniraum_mapper_ogg_deinit(uniraum_mapping_t * mapping) {
 struct ogg_mapping * self = mapping->state.vp;

 if ( self == NULL )
  return 0;

 if ( self->oy_opened )
  ogg_sync_clear(&(self->oy));

 if ( self != NULL )
  roar_mm_free(self);

 return 0;
}

static uniraum_data_t * _req2page(uniraum_mapping_t * mapping) {
 struct ogg_mapping * self = mapping->state.vp;
 uniraum_data_t * page;
 uniraum_req_t  * req;
 ogg_page og;
 ogg_page * ogp;
 void * pagedata;
 char * syncbuf;
 int reqtype;
 int reqlen;
 void * reqdata;

 if ( mapping->req_queue_fill == 0 )
  return NULL;

 req     = mapping->req_queue[0];

 reqtype = uniraum_reqget_type(req);
 reqlen  = uniraum_reqget_datalen(req);
 reqdata = uniraum_reqget_data(req);

 // this test shpuld be in a loop...
 if ( reqtype != ROAR_CMD_ADD_DATA ) {
  uniraum_requnref(uniraum_mapping_shiftreq(mapping));
  return NULL;
 }

 if ( reqlen == 0 ) {
  // update typesstamps here.
  uniraum_requnref(uniraum_mapping_shiftreq(mapping));
  return NULL;
 }

 if ( !self->oy_opened ) {
  ogg_sync_init(&(self->oy));
  self->oy_opened = 1;
 }

 syncbuf = ogg_sync_buffer(&(self->oy), reqlen);
 memcpy(syncbuf, reqdata, reqlen);
 ogg_sync_wrote(&(self->oy), reqlen);

 uniraum_requnref(uniraum_mapping_shiftreq(mapping));

 if ( ogg_sync_pageout(&(self->oy), &og) < 1 )
  return NULL;

 page = uniraum_data_new(sizeof(og)+og.header_len+og.body_len, &pagedata, ROAR_VIO_DFT_OGG_PAGE);
 if ( page == NULL )
  return NULL;

 memcpy(pagedata, &og, sizeof(og));
 memcpy(pagedata+sizeof(og), og.header, og.header_len);
 memcpy(pagedata+sizeof(og)+og.header_len, og.body, og.body_len);

 ogp = pagedata;

 // update pointers:
 ogp->header = pagedata+sizeof(og);
 ogp->body   = pagedata+sizeof(og)+og.header_len;

 return page;
}

int uniraum_mapper_ogg_req2page(uniraum_mapping_t * mapping) {
 uniraum_data_t * page;

 if ( mapping->req_queue_fill == 0 )
  return 0;

 while (mapping->data_queue_fill != _DATA_QUEUE_LEN) {
  page = _req2page(mapping);
  if ( page == NULL ) {
   break;
  } else {
   uniraum_mapping_pushdata(mapping, page);
   roar_buffer_unref(page);
  }
 }

 return 0;
}

int uniraum_mapper_ogg_req2raw(uniraum_mapping_t * mapping) {
 uniraum_data_t * raw;
 uniraum_req_t  * req;
 void           * rawdata;
 int              reqtype;
 int              reqlen;
 void           * reqdata;

 if ( mapping->req_queue_fill == 0 )
  return 0;

 while (mapping->data_queue_fill != _DATA_QUEUE_LEN) {
  if ( mapping->req_queue_fill == 0 )
   return 0;

  req     = mapping->req_queue[0];

  reqtype = uniraum_reqget_type(req);
  reqlen  = uniraum_reqget_datalen(req);
  reqdata = uniraum_reqget_data(req);

  // this test shpuld be in a loop...
  if ( reqtype != ROAR_CMD_ADD_DATA ) {
   uniraum_requnref(uniraum_mapping_shiftreq(mapping));
   return -1;
  }

  if ( reqlen == 0 ) {
   // update typesstamps here.
   uniraum_requnref(uniraum_mapping_shiftreq(mapping));
   return -1;
  }

  raw = uniraum_data_new(reqlen, &rawdata, ROAR_VIO_DFT_RAW);
  if ( raw == NULL )
   return 0;

  uniraum_requnref(uniraum_mapping_shiftreq(mapping));

  memcpy(rawdata, reqdata, reqlen);

  uniraum_mapping_pushdata(mapping, raw);
  roar_buffer_unref(raw);
 }

 return 0;
}

static int _page2req (uniraum_mapping_t * mapping, ogg_page * page) {
 ogg_int64_t page_pos = ogg_page_granulepos(page);
 uniraum_req_t * req = uniraum_reqdata(mapping->stream, mapping->state.pos, NULL, page->header_len + page->body_len);
 void * data;

 if ( req == NULL )
  return -1;

 data = uniraum_reqget_data(req);
 memcpy(data, page->header, page->header_len);
 memcpy(data+page->header_len, page->body, page->body_len);

 uniraum_mapping_pushreq(mapping, req);
 uniraum_requnref(req);

 if ( page_pos > (ogg_int64_t)mapping->state.pos )
  mapping->state.pos = page_pos;

 return 0;
}

int uniraum_mapper_ogg_page2req(uniraum_mapping_t * mapping) {
 uniraum_data_t * buf;
 size_t len;
 void * data;
 int type;

 while ( mapping->data_queue_fill != 0 && mapping->req_queue_fill != _REQ_QUEUE_LEN ) {
  buf = mapping->data_queue[0];

  if ( roar_buffer_get_len(buf, &len) == -1 )
   return -1;

  if ( len < sizeof(ogg_page) )
   return -1;

  if ( roar_buffer_get_data(buf, &data) == -1 )
   return -1;

  if ( roar_buffer_get_type(buf, &type) == -1 )
   return -1;

  if ( type != ROAR_VIO_DFT_OGG_PAGE )
   return -1;

  if ( _page2req(mapping, data) == -1 )
   return -1;

  uniraum_data_unref(uniraum_mapping_shiftdata(mapping));
 }

 return 0;
}

int uniraum_mapper_ogg_raw2req(uniraum_mapping_t * mapping) {
 struct ogg_mapping * self = mapping->state.vp;
 uniraum_data_t * buf;
 ogg_page page;
 size_t len;
 void * data;
 int type;
 char * syncbuf;

 while ( mapping->data_queue_fill != 0 && mapping->req_queue_fill != _REQ_QUEUE_LEN ) {
  buf = mapping->data_queue[0];

  if ( roar_buffer_get_len(buf, &len) == -1 )
   return -1;

  if ( roar_buffer_get_data(buf, &data) == -1 )
   return -1;

  if ( roar_buffer_get_type(buf, &type) == -1 )
   return -1;

  if ( type != ROAR_VIO_DFT_RAW && type != ROAR_VIO_DFT_PACKET )
   return -1;

  if ( len == 0 ) {
   // update typesstamps here.
   uniraum_data_unref(uniraum_mapping_shiftdata(mapping));
   continue;
  }

  if ( !self->oy_opened ) {
   ogg_sync_init(&(self->oy));
   self->oy_opened = 1;
  }

  syncbuf = ogg_sync_buffer(&(self->oy), len);
  memcpy(syncbuf, data, len);
  ogg_sync_wrote(&(self->oy), len);

  uniraum_data_unref(uniraum_mapping_shiftdata(mapping));
 }

 while (mapping->req_queue_fill != _REQ_QUEUE_LEN && ogg_sync_pageout(&(self->oy), &page) > 0) {
  _page2req(mapping, &page);
 }

 return 0;
}

#endif

//ll
