//mapping_pcm.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "private.h"

int uniraum_mapper_pcm_init(uniraum_mapping_t * mapping) {
 ssize_t          bpf    = roar_info2framesize(&(mapping->info));

 if ( bpf == -1 ) {
  if ( mapping->info.codec == ROAR_CODEC_DMX512 ) {
   if ( mapping->info.channels == 0 )
    mapping->info.channels = 512;

   bpf = 8*mapping->info.channels;
  } else {
   return -1;
  }
 }

 mapping->state.bpf = bpf;

 return 0;
}

int uniraum_mapper_pcm_deinit(uniraum_mapping_t * mapping) {
 (void)mapping;
 return 0;
}

int uniraum_mapper_pcm_req2packet(uniraum_mapping_t * mapping) {
 uniraum_data_t * buf = NULL;
 uniraum_req_t  * req;
 int type;
 ssize_t len;
 void * data, *bufdata;
 int err;

 if ( mapping->req_queue_fill == 0 || mapping->data_queue_fill == _DATA_QUEUE_LEN )
  return 0;

 req  = mapping->req_queue[0];
 type = uniraum_reqget_type(req);

 if ( type == ROAR_CMD_ADD_DATA ) {
  len  = uniraum_reqget_datalen(req);
  data = uniraum_reqget_data(req);

  if ( roar_buffer_new_data(&buf, len, &bufdata) == -1 )
   return -1;

  if ( roar_buffer_set_type(buf, mapping->type) == -1 ) {
   err = roar_error;
   roar_buffer_unref(buf);
   roar_err_set(err);
   return -1;
  }

  memcpy(bufdata, data, len);

  if ( uniraum_mapping_pushdata(mapping, buf) == -1 ) {
   err = roar_error;
   roar_buffer_unref(buf);
   roar_err_set(err);
   return -1;
  }
 }

 // cleanup:
 (void)uniraum_mapping_shiftreq(mapping);

 if ( buf != NULL )
  roar_buffer_unref(buf);
 uniraum_requnref(req);

 return 0;
}

int uniraum_mapper_pcm_raw2req(uniraum_mapping_t * mapping) {
 uniraum_data_t * buf_new;
 uniraum_data_t * buf;
 ssize_t buflen;
 ssize_t statelen;
 size_t clen, rlen;
 void * bufdata;
 void * buf_newdata;

 ROAR_DBG("uniraum_mapper_pcm_raw2req(mapping=%p) = ?", mapping);

 if ( mapping->data_queue_fill == 0 || mapping->req_queue_fill == _REQ_QUEUE_LEN )
  return 0;

 buf = mapping->data_queue[0];

 if ( mapping->state.buf == NULL ) {
  statelen = 0;
 } else {
  statelen = uniraum_data_get_len(mapping->state.buf);
 }

 if ( statelen == -1 ) {
  return -1;
 }

 buflen = uniraum_data_get_len(buf);

 if ( buflen == -1 ) {
  return -1;
 }

 clen  = statelen + buflen;

 ROAR_DBG("uniraum_mapper_pcm_raw2req(mapping=%p): statelen=%llu, buflen=%llu", mapping,
               (long long unsigned int)statelen, (long long unsigned int)buflen);

 clen /= mapping->state.bpf / (size_t)8;

 if ( clen == 0 ) {
  clen  = statelen + buflen;

  if ( roar_buffer_new_data(&buf_new, clen, &buf_newdata) == -1 )
   return -1;

  if ( statelen ) {
   bufdata = uniraum_data_get_data(mapping->state.buf);
   memcpy(buf_newdata, bufdata, statelen);
  }

  bufdata = uniraum_data_get_data(buf);
  memcpy(buf_newdata+statelen, bufdata, buflen);

  roar_buffer_unref(mapping->state.buf);
  mapping->state.buf = buf_new;

  uniraum_data_unref(uniraum_mapping_shiftdata(mapping));

  return 0;
 }

 clen  *= mapping->state.bpf / (size_t)8;

 if ( roar_buffer_new_data(&buf_new, clen, &buf_newdata) == -1 )
  return -1;

 if ( roar_buffer_set_type(buf_new, ROAR_VIO_DFT_UNFRAMED) == -1 )
  return -1;

 if ( statelen ) {
  bufdata = uniraum_data_get_data(mapping->state.buf);
  memcpy(buf_newdata, bufdata, statelen);
  roar_buffer_unref(mapping->state.buf);
  mapping->state.buf = NULL;
 }

 bufdata = uniraum_data_get_data(buf);
 memcpy(buf_newdata+statelen, bufdata, clen-statelen);

 mapping->data_queue[0] = buf_new;

 if ( uniraum_mapper_pcm_packet2req(mapping) == -1 ) {
  ROAR_DBG("uniraum_mapper_pcm_raw2req(mapping=%p): uniraum_mapper_pcm_packet2req(mapping=%p) returned error: %s", mapping, mapping, roar_error2str(uniraum_error(NULL)));
 }

 rlen  = statelen + buflen;

 rlen %= mapping->state.bpf / (size_t)8;

 if ( rlen ) {
  if ( roar_buffer_new_data(&buf_new, rlen, &buf_newdata) == -1 )
   return -1;
  memcpy(buf_newdata, bufdata+buflen-rlen, rlen);
  mapping->state.buf = buf_new;
 }

 ROAR_DBG("uniraum_mapper_pcm_raw2req(mapping=%p) = 0", mapping);
 return 0;
}

int uniraum_mapper_pcm_packet2req(uniraum_mapping_t * mapping) {
 uniraum_data_t * buf;
 uniraum_req_t  * req;
 size_t len;
 void * data;
 int err;
 int type;

 ROAR_DBG("uniraum_mapper_pcm_packet2req(mapping=%p) = ?", mapping);

 if ( mapping->data_queue_fill == 0 || mapping->req_queue_fill == _REQ_QUEUE_LEN )
  return 0;

 buf = mapping->data_queue[0];

 if ( roar_buffer_get_len(buf, &len) == -1 )
  return -1;

 if ( roar_buffer_get_data(buf, &data) == -1 )
  return -1;

 if ( roar_buffer_get_type(buf, &type) == -1 )
  return -1;

 switch (type) {
  case ROAR_VIO_DFT_PACKET:
  case ROAR_VIO_DFT_UNFRAMED:
    // ok, we can use those.
   break;
  default:
    roar_err_set(ROAR_ERROR_TYPEMM);
    return -1;
 }

 req =  uniraum_reqdata(mapping->stream, mapping->state.pos, data, len);
 if ( req == NULL )
  return -1;

 if ( uniraum_mapping_pushreq(mapping, req) == -1 ) {
  err = roar_error;
  uniraum_requnref(req);
  roar_err_set(err);
 }

 uniraum_requnref(req);

 mapping->state.iolen += len;
 mapping->state.pos    = (mapping->state.iolen * (size_t)8) / mapping->state.bpf;

 uniraum_data_unref(uniraum_mapping_shiftdata(mapping));

 ROAR_DBG("uniraum_mapper_pcm_packet2req(mapping=%p) = 0", mapping);
 return 0;
}

//ll
