//mapping_pcm.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "private.h"

#ifdef HAVE_LIBOGG

#include <ogg/ogg.h>

#define PS_OPTIMAL        4095
#define PS_MINSPLIT       6143
#define PS_MINSPLIT_LAST  5389
#define PS_MAXSIZE       65534

#define UR_OGG_BOS    0x01
#define UR_OGG_EOS    0x02
#define UR_OGG_CONT   0x04
#define UR_OGG_LONG   0x08
#define UR_OGG_MULTI  0x10

struct xiph_mapping {
 ogg_stream_state os;
 int os_opened;
 ogg_sync_state oy;
 int oy_opened;
 uniraum_data_t * pkginbuf;
 size_t pkginbuf_avg_len;
};

static int _packet2req(uniraum_mapping_t * mapping, ogg_packet * pkg);
static int _page2req(uniraum_mapping_t * mapping, ogg_page * page);


static size_t _reqsize (size_t pkglen, size_t max_num) {
 size_t num;
 size_t guess0, guess1;

 if ( pkglen < PS_MINSPLIT )
  return pkglen;

 num = pkglen / PS_OPTIMAL;

 if ( max_num && num >= max_num ) {
  num = max_num - 1;
 }

 guess0 = pkglen / (num + (size_t)1);
 guess1 = pkglen / num;

 if ( labs(PS_OPTIMAL - guess0) < labs(PS_OPTIMAL - guess1) ) {
  return guess0;
 } else {
  return guess1;
 }
}

int uniraum_mapper_xiph_init(uniraum_mapping_t * mapping) {
 struct xiph_mapping * self = roar_mm_malloc(sizeof(struct xiph_mapping));

 if ( self == NULL )
  return -1;

 memset(self, 0, sizeof(struct xiph_mapping));
 self->pkginbuf = NULL;

 mapping->state.vp = self;

 return 0;
}

int uniraum_mapper_xiph_deinit(uniraum_mapping_t * mapping) {
 struct xiph_mapping * self = mapping->state.vp;

 if ( self == NULL )
  return 0;

 if ( self->os_opened )
  ogg_stream_clear(&(self->os));

 if ( self->oy_opened )
  ogg_sync_clear(&(self->oy));

 if ( self->pkginbuf != NULL )
  uniraum_data_unref(self->pkginbuf);

 if ( self != NULL )
  roar_mm_free(self);

 return 0;
}

static uniraum_data_t * _req2packet(uniraum_mapping_t * mapping) {
 struct xiph_mapping * self = mapping->state.vp;
 uniraum_data_t * pkg = self->pkginbuf;
 uniraum_req_t  * req;
 ogg_packet * op = NULL;
 size_t buflen, reqlen;
 void * bufdata;
 unsigned char * reqdata;
 int err;
 int reqtype;

 if ( pkg != NULL )
  op = uniraum_data_get_data(pkg);

 ROAR_DBG("_req2packet(*) = ?");

 while (mapping->req_queue_fill != 0) {
  ROAR_DBG("_req2packet(*) = ?");

  req     = mapping->req_queue[0];
  reqlen  = uniraum_reqget_datalen(req);
  reqdata = uniraum_reqget_data(req);
  reqtype = uniraum_reqget_type(req);

  ROAR_DBG("_req2packet(*): req=%p{.len=%lli, .data=%p, .type=%i}", req, (long long signed int)(ssize_t)reqlen, reqdata, reqtype);

  if ( reqtype != ROAR_CMD_ADD_DATA ) {
   uniraum_requnref(uniraum_mapping_shiftreq(mapping));
   continue;
  }

  if ( reqlen == 0 ) {
   // update typesstamps here.
   uniraum_requnref(uniraum_mapping_shiftreq(mapping));
   continue;
  }

//  printf("req={type=%i, len=%zi, data=%p}\n", reqtype, reqlen, reqdata);

  if ( pkg == NULL ) {
   buflen = self->pkginbuf_avg_len;
   ROAR_DBG("_req2packet(*): buflen=%llu", (long long unsigned int)buflen);

   if ( reqdata[0] & UR_OGG_LONG ) {
    if ( buflen < (((reqlen * 3) / 2) + sizeof(ogg_packet)) )
     buflen = reqlen * 2 + sizeof(ogg_packet);
   } else {
    if ( buflen < (reqlen + sizeof(ogg_packet)) )
     buflen = reqlen + sizeof(ogg_packet);
   }

   ROAR_DBG("_req2packet(*): buflen=%llu", (long long unsigned int)buflen);

   if ( roar_buffer_new(&pkg, buflen) == -1 )
    return NULL;

   ROAR_DBG("_req2packet(*) = ?");

   if ( roar_buffer_set_type(pkg, ROAR_VIO_DFT_OGG_PACKET) == -1 ) {
    err = roar_error;
    roar_buffer_unref(pkg);
    roar_err_set(err);
    return NULL;
   }

   ROAR_DBG("_req2packet(*) = ?");

   if ( roar_buffer_set_len(pkg, sizeof(ogg_packet)) == -1 ) {
    err = roar_error;
    roar_buffer_free(pkg);
    roar_err_set(err);
    return NULL;
   }

   ROAR_DBG("_req2packet(*) = ?");

   bufdata = uniraum_data_get_data(pkg);

   self->pkginbuf = pkg;
   op             = bufdata;

   memset(op, 0, sizeof(ogg_packet));
  }

  if ( reqdata[0] & UR_OGG_BOS )
   op->b_o_s = 1;
  if ( reqdata[0] & UR_OGG_EOS )
   op->e_o_s = 1;

  bufdata = uniraum_data_get_data(pkg);
  buflen = uniraum_data_get_len(pkg);
  ROAR_DBG("_req2packet(*): pkg=%p{.len=%lli, .data=%p}", pkg, (long long signed int)(ssize_t)buflen, bufdata);

  if ( roar_buffer_set_len(pkg, buflen + reqlen - 1) == -1 ) {
   return NULL;
  }

  op = bufdata = uniraum_data_get_data(pkg);

  ROAR_DBG("_req2packet(*): pkg=%p{.len=%lli, .data=%p}", pkg, (long long signed int)(ssize_t)(buflen + reqlen - 1), bufdata);
  ROAR_DBG("_req2packet(*): memcpy(%p, %p, %lli)", bufdata + buflen, reqdata + 1, (long long signed int)((ssize_t)reqlen - (ssize_t)1));
  memcpy(bufdata + buflen, reqdata + 1, reqlen - 1);

  if ( !(reqdata[0] & UR_OGG_LONG) ) {
   self->pkginbuf = NULL;
   op->packet     = bufdata                   + sizeof(ogg_packet);
   op->bytes      = uniraum_data_get_len(pkg) - sizeof(ogg_packet);
   op->granulepos = -1; // TODO: fix this later.
   op->packetno   =  0;
  }

  ROAR_DBG("_req2packet(*) = ?");
  uniraum_requnref(uniraum_mapping_shiftreq(mapping));
 }

 if ( self->pkginbuf == NULL && pkg != NULL ) {
  buflen = uniraum_data_get_len(pkg);
  if ( self->pkginbuf_avg_len == 0 ) {
   // TODO: tune this later as initial packets tend to be small headers.
   self->pkginbuf_avg_len = buflen;
  } else {
   self->pkginbuf_avg_len = (self->pkginbuf_avg_len*8 + buflen*2) / 10;
  }
  return pkg;
 }

 ROAR_DBG("_req2packet(*) = NULL");
 return NULL;
}

int uniraum_mapper_xiph_req2packet(uniraum_mapping_t * mapping) {
 uniraum_data_t * pkg;

 if ( mapping->req_queue_fill == 0 )
  return 0;

 while (mapping->data_queue_fill != _DATA_QUEUE_LEN) {
  pkg = _req2packet(mapping);
  if ( pkg == NULL ) {
   break;
  } else {
   uniraum_mapping_pushdata(mapping, pkg);
   roar_buffer_unref(pkg);
  }
 }

 return 0;
}

static uniraum_data_t * _req2page(uniraum_mapping_t * mapping) {
 struct xiph_mapping * self = mapping->state.vp;
 uniraum_data_t * page;
 uniraum_data_t * pkg;
 ogg_packet * op;
 ogg_page og;
 ogg_page * ogp;
 void * pagedata;
 int err;

 if ( !self->os_opened ) {
  ogg_stream_init(&(self->os), mapping->stream);
  self->os_opened = 1;
 }

 while (ogg_stream_pageout(&(self->os), &og) < 1) {
  if ( (pkg = _req2packet(mapping)) == NULL )
   return NULL;

  op = uniraum_data_get_data(pkg);

  ogg_stream_packetin(&(self->os), op);

  uniraum_data_unref(pkg);
 }

 if ( roar_buffer_new_data(&page, og.header_len + og.body_len + sizeof(ogg_page), &pagedata) == -1 )
  return NULL;

 if ( roar_buffer_set_type(page, ROAR_VIO_DFT_OGG_PAGE) == -1 ) {
  err = roar_error;
  roar_buffer_unref(page);
  roar_err_set(err);
  return NULL;
 }

 memcpy(pagedata, &og, sizeof(og));
 memcpy(pagedata+sizeof(og), og.header, og.header_len);
 memcpy(pagedata+sizeof(og)+og.header_len, og.body, og.body_len);

 ogp = pagedata;

 // update pointers:
 ogp->header = pagedata+sizeof(og);
 ogp->body   = pagedata+sizeof(og)+og.header_len;

 return page;
}

int uniraum_mapper_xiph_req2page(uniraum_mapping_t * mapping) {
 uniraum_data_t * page;

 if ( mapping->req_queue_fill == 0 )
  return 0;

 while (mapping->data_queue_fill != _DATA_QUEUE_LEN) {
  page = _req2page(mapping);
  if ( page == NULL ) {
   break;
  } else {
   uniraum_mapping_pushdata(mapping, page);
   roar_buffer_unref(page);
  }
 }

 return 0;
}

int uniraum_mapper_xiph_req2raw(uniraum_mapping_t * mapping) {
 uniraum_data_t * page;
 uniraum_data_t * raw;
 ogg_page       * og;
 void           * rawdata;
 int              err;

 if ( mapping->req_queue_fill == 0 )
  return 0;

 while (mapping->data_queue_fill != _DATA_QUEUE_LEN) {
  page = _req2page(mapping);
  if ( page == NULL ) {
   break;
  } else {
   og = uniraum_data_get_data(page);

   if ( roar_buffer_new_data(&raw, og->header_len + og->body_len, &rawdata) == -1 ) {
    uniraum_data_unref(page);
    return -1;
   }

   if ( roar_buffer_set_type(raw, mapping->type) == -1 ) {
    err = roar_error;
    roar_buffer_unref(raw);
    roar_err_set(err);
    return -1;
   }

   memcpy(rawdata,                og->header, og->header_len);
   memcpy(rawdata+og->header_len, og->body,   og->body_len);

   uniraum_data_unref(page);

   uniraum_mapping_pushdata(mapping, raw);
   roar_buffer_unref(raw);
  }
 }

 return 0;
}

int uniraum_mapper_xiph_page2req(uniraum_mapping_t * mapping) {
 uniraum_data_t * buf;
 size_t len;
 void * data;
 int type;

 if ( mapping->req_queue_fill == _REQ_QUEUE_LEN )
  return 0;

 if ( mapping->data_queue_fill != 0 ) {
  buf = mapping->data_queue[0];

  if ( roar_buffer_get_len(buf, &len) == -1 )
   return -1;

  if ( len < sizeof(ogg_page) )
   return -1;

  if ( roar_buffer_get_data(buf, &data) == -1 )
   return -1;

  if ( roar_buffer_get_type(buf, &type) == -1 )
   return -1;

  if ( type != ROAR_VIO_DFT_OGG_PAGE )
   return -1;

  if ( _page2req(mapping, data) == -1 )
   return -1;

  uniraum_data_unref(uniraum_mapping_shiftdata(mapping));
 } else {
  _page2req(mapping, NULL);
 }

 return 0;
}

static int _page2req(uniraum_mapping_t * mapping, ogg_page * page) {
 struct xiph_mapping * self = mapping->state.vp;
 ogg_packet pkg;

 if ( !self->os_opened ) {
  if ( page == NULL )
   return 0;

  ogg_stream_init(&(self->os), ogg_page_serialno(page));
  self->os_opened = 1;
 }

 if ( page != NULL ) {
  if ( self->os.serialno != ogg_page_serialno(page) )
   return -1;

  ogg_stream_pagein(&(self->os), page);
 }

 while (mapping->req_queue_fill != _REQ_QUEUE_LEN && ogg_stream_packetout(&(self->os), &pkg) > 0) {
  _packet2req(mapping, &pkg);
 }

 return 0;
}

int uniraum_mapper_xiph_raw2req(uniraum_mapping_t * mapping) {
 struct xiph_mapping * self = mapping->state.vp;
 uniraum_data_t * buf;
 ogg_page page;
 size_t len;
 void * data;
 int type;
 char * syncbuf;

 if ( mapping->req_queue_fill == _REQ_QUEUE_LEN )
  return 0;

 if ( mapping->data_queue_fill != 0 ) {
  buf = mapping->data_queue[0];

  if ( roar_buffer_get_len(buf, &len) == -1 )
   return -1;

  if ( roar_buffer_get_data(buf, &data) == -1 )
   return -1;

  if ( roar_buffer_get_type(buf, &type) == -1 )
   return -1;

  if ( type != ROAR_VIO_DFT_RAW && type != ROAR_VIO_DFT_PACKET )
   return -1;

  if ( !self->oy_opened ) {
   ogg_sync_init(&(self->oy));
   self->oy_opened = 1;
  }

  syncbuf = ogg_sync_buffer(&(self->oy), len);
  memcpy(syncbuf, data, len);
  ogg_sync_wrote(&(self->oy), len);

  uniraum_data_unref(uniraum_mapping_shiftdata(mapping));
 }

 while (mapping->req_queue_fill != _REQ_QUEUE_LEN && ogg_sync_pageout(&(self->oy), &page) > 0) {
  _page2req(mapping, &page);
 }

 _page2req(mapping, NULL); // flush buffers

 return 0;
}

int uniraum_mapper_xiph_packet2req(uniraum_mapping_t * mapping) {
 uniraum_data_t * buf;
 size_t len;
 void * data;
 int type;

 if ( mapping->data_queue_fill == 0 || mapping->req_queue_fill == _REQ_QUEUE_LEN )
  return 0;

 buf = mapping->data_queue[0];

 if ( roar_buffer_get_len(buf, &len) == -1 )
  return -1;

 if ( len < sizeof(ogg_packet) )
  return -1;

 if ( roar_buffer_get_data(buf, &data) == -1 )
  return -1;

 if ( roar_buffer_get_type(buf, &type) == -1 )
  return -1;

 if ( _packet2req(mapping, data) == -1 )
  return -1;

 uniraum_data_unref(uniraum_mapping_shiftdata(mapping));

 return 0;
}

static int _packet2req(uniraum_mapping_t * mapping, ogg_packet * pkg) {
 uniraum_req_t * req;
 size_t len, reqlen;
 size_t todo, done;
 void * data;
 unsigned char flags;
 int err;

 if ( pkg->packet == NULL )
  return -1;

 len = pkg->bytes;

 reqlen = _reqsize(len, _REQ_QUEUE_LEN - mapping->req_queue_fill);
 if ( reqlen >= PS_MAXSIZE ) {
  roar_err_set(ROAR_ERROR_MSGSIZE);
  return -1;
 }

 todo = len;
 done = 0;

 while (todo) {
  flags = 0;

  if ( pkg->b_o_s )
   flags |= UR_OGG_BOS;

  if ( pkg->e_o_s )
   flags |= UR_OGG_EOS;

  if ( reqlen > todo ) {
   reqlen = todo;
  } else if ( todo > reqlen && !(todo > PS_MINSPLIT_LAST) ) {
   reqlen = todo;
  }

  if ( todo > reqlen )
   flags |= UR_OGG_LONG;

  req = uniraum_reqdata(mapping->stream, mapping->state.pos, NULL, reqlen+1);
  data = uniraum_reqget_data(req);

  *(unsigned char*)data = flags;

  memcpy(data+1, pkg->packet+done, reqlen);

  if ( uniraum_mapping_pushreq(mapping, req) == -1 ) {
   err = roar_error;
   uniraum_requnref(req);
   roar_err_set(err);
   return -1;
  }

  uniraum_requnref(req);

  done += reqlen;
  todo -= reqlen;
 }

 if ( pkg->granulepos > (int64_t)mapping->state.pos )
  mapping->state.pos = pkg->granulepos - 1;

 return 0;
}

#endif

//ll
