//req.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "private.h"

#define _ckreq(r) if ( (r) == NULL ) { roar_error = ROAR_ERROR_FAULT; return -1; }

uniraum_req_t *  uniraum_reqnewempty(void) {
 uniraum_req_t * req = roar_mm_malloc(sizeof(uniraum_req_t));

 if (req == NULL) {
  return NULL;
 }

 memset(req, 0, sizeof(uniraum_req_t));

 req->refc       =  1;
 req->mes.stream = -1;
 req->long_data  = NULL;
 req->buf        = NULL;

 return req;
}

int              uniraum_reqref(uniraum_req_t * req) {
 _ckreq(req);

 req->refc++;

 return 0;
}

int              uniraum_requnref(uniraum_req_t * req) {
 _ckreq(req);

 req->refc--;

 if ( req->refc == 0 ) {
  ROAR_DBG("uniraum_requnref(req=%p{.refc = 0, .mes = {...}, .long_data=%p}) = ?", req, req->long_data);
  if ( req->long_data != NULL )
   free(req->long_data);
  if ( req->buf != NULL )
   free(req->buf);
  roar_mm_free(req);
 }

 return 0;
}

uniraum_req_t *  uniraum_reqexec(int stream, uniraum_pos_t pos) {
 uniraum_req_t * ret = uniraum_reqnewempty();

 if ( ret == NULL )
  return NULL;

 ret->mes.cmd    = ROAR_CMD_EXEC_STREAM;
 ret->mes.stream = stream;
 ret->mes.pos    = pos;

 return ret;
}

uniraum_req_t *  uniraum_reqnoop(size_t len, int stream, uniraum_pos_t pos, int random) {
 uniraum_req_t * ret = uniraum_reqnewempty();
 void * data;

 if ( ret == NULL )
  return NULL;

 ret->mes.cmd     = ROAR_CMD_NOOP;
 ret->mes.stream  = stream;
 ret->mes.pos     = pos;
 ret->mes.datalen = len;

 if ( len > sizeof(ret->mes.data) ) {
  data = malloc(len);
  if ( data == NULL ) {
   uniraum_requnref(ret);
   return NULL;
  }
  ret->long_data = data;
 } else {
  data = ret->mes.data;
 }

 if ( random ) {
  if ( roar_random_gen_nonce(data, len) == -1 ) {
   uniraum_requnref(ret);
   return NULL;
  }
 } else {
  memset(data, 0, len);
 }

 return ret;
}

uniraum_req_t *  uniraum_reqpadding(size_t len) {
 return uniraum_reqnoop(len, -1, 0, 0);
}

uniraum_req_t *  uniraum_reqeof(void) {
 uniraum_req_t * ret = uniraum_reqnewempty();

 if ( ret == NULL )
  return NULL;

 ret->mes.cmd    = ROAR_CMD_QUIT;

 return ret;
}

uniraum_req_t *  uniraum_reqnew_stream(int stream, int parent, int dir, struct roar_audio_info * info) {
 struct roar_stream  ms;
 uniraum_req_t * ret;

 if ( info == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return NULL;
 }

 ret = uniraum_reqnewempty();

 if ( ret == NULL )
  return NULL;

 memset(&ms,  0, sizeof(ms));

 ms.dir        = dir;
 ms.pos_rel_id = parent;
 ms.pos        = 0;

 memcpy(&(ms.info), info, sizeof(ms.info));

 roar_stream_s2m(&ms, &(ret->mes));

 ret->mes.cmd    = ROAR_CMD_NEW_STREAM;
 ret->mes.stream = stream;

 return ret;
}

uniraum_req_t *  uniraum_reqnew_stream_simple(int * stream, uint_least32_t rate, uint_least32_t bits, uint_least32_t channels, uint_least32_t codec) {
 struct roar_audio_info info;
 uint16_t stream_id = roar_random_uint16() & 0x7FFF;

 if ( stream == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return NULL;
 }

 if ( rate == 0 || bits == 0 || channels == 0 || codec == 0 ) {
  roar_error = ROAR_ERROR_INVAL;
  return NULL;
 }

 info.rate     = rate;
 info.bits     = bits;
 info.channels = channels;
 info.codec    = codec;

 *stream       = stream_id;

 return uniraum_reqnew_stream(stream_id, -1, ROAR_DIR_PLAY, &info);
}

uniraum_req_t *  uniraum_reqend_stream(int stream, uniraum_pos_t pos) {
 uniraum_req_t * ret = uniraum_reqnewempty();
 uint16_t * info;

 if ( ret == NULL )
  return NULL;

 info = (uint16_t*)ret->mes.data;

 ret->mes.cmd     = ROAR_CMD_KICK;
 ret->mes.stream  = stream;
 ret->mes.pos     = pos;
 ret->mes.datalen = 4;

 info[0] = ROAR_HOST2NET16(ROAR_OT_STREAM);
 info[1] = ROAR_HOST2NET16(stream);

 return ret;
}

static uniraum_req_t * _req_meta(int stream, const char * key, const char * value, int mode) {
 uniraum_req_t * ret;
 size_t len;
 int type;

 if ( key == NULL ) {
  type = ROAR_META_TYPE_NONE;
 } else {
  type = roar_meta_inttype(key);

  if ( type == -1 )
   return NULL;
 }

 if ( value == NULL )
  value = "";

 ret = uniraum_reqnewempty();

 if ( ret == NULL )
  return NULL;

 ret->mes.cmd    = ROAR_CMD_SET_META;
 ret->mes.stream = stream;

 ret->mes.data[0] = 0;
 ret->mes.data[1] = mode;
 ret->mes.data[2] = type;

 ret->mes.data[3] = 0;
 ret->mes.data[4] = len = strlen(value);

 if ( len > 255 ) {
  uniraum_requnref(ret);
  roar_error = ROAR_ERROR_MSGSIZE;
  return NULL;
 }

 ret->mes.datalen = 5 + ret->mes.data[3] + ret->mes.data[4];

 if ( ret->mes.datalen > (ssize_t)sizeof(ret->mes.data) ) {
  uniraum_requnref(ret);
  roar_error = ROAR_ERROR_MSGSIZE;
  return NULL;
 }

 strncpy(&(ret->mes.data[5+ret->mes.data[3]]), value, len);

 return ret;
}

uniraum_req_t *  uniraum_reqset_meta(int stream, const char * key, const char * value) {
 return _req_meta(stream, key, value, ROAR_META_MODE_SET);
}

uniraum_req_t *  uniraum_reqadd_meta(int stream, const char * key, const char * value) {
 return _req_meta(stream, key, value, ROAR_META_MODE_ADD);
}

uniraum_req_t *  uniraum_reqfin_meta(int stream) {
 return _req_meta(stream, NULL, NULL, ROAR_META_MODE_FINALIZE);
}

uniraum_req_t *  uniraum_reqclear_meta(int stream) {
 return _req_meta(stream, NULL, NULL, ROAR_META_MODE_CLEAR);
}


uniraum_req_t *  uniraum_reqdata(int stream, uniraum_pos_t pos, const void * data, size_t len) {
 uniraum_req_t * ret = uniraum_reqnewempty();
 void * buf = NULL;
 int err;

 if ( ret == NULL )
  return NULL;

 if ( len <= sizeof(ret->mes.data) ) {
  if ( data != NULL ) {
   memcpy(ret->mes.data, data, len);
  }
 } else {
  if ( data == NULL ) {
   buf = roar_mm_malloc(len);
   if ( buf != NULL )
    memset(buf, 0, len);
  } else {
   buf = roar_mm_memdup(data, len);
  }

  if ( buf == NULL ) {
   err = roar_error;
   uniraum_requnref(ret);
   roar_err_set(err);
   return NULL;
  }
 }

 ret->mes.cmd     = ROAR_CMD_ADD_DATA;
 ret->mes.datalen = len;
 ret->mes.stream  = stream;
 ret->mes.pos     = pos;

 ret->long_data   = buf;

 return ret;
}

uniraum_req_t *  uniraum_reqvolume(int stream, ...);

uniraum_req_t *  uniraum_reqstream_para(int stream, ...);

uniraum_req_t *  uniraum_reqcaps(uniraum_cap_t * caps0, size_t len0, uniraum_cap_t * caps1, size_t len1) {
 uniraum_req_t * ret = NULL;
 struct roar_caps caps;
 struct roar_stds * stds = NULL;
 int err;

 if ( (caps0 == NULL && len0 != 0) || (caps1 == NULL && len1 != 0) ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 if ( len0 == 0 && len1 == 0 ) {
  return uniraum_reqpadding(0);
 }

 if ( (stds = roar_stds_new(len0 + len1)) == NULL )
  return NULL;

 if ( len0 != 0 )
  memcpy(stds->stds, caps0, len0*sizeof(uniraum_cap_t));
 if ( len1 != 0 )
  memcpy(stds->stds+len0*sizeof(uniraum_cap_t), caps1, len1*sizeof(uniraum_cap_t));

 memset(&caps, 0, sizeof(caps));
 caps.version = 0;
 caps.type    = ROAR_CT_STANDARDS;
 caps.flags   = 0;
 caps.data    = stds->stds;
 caps.len     = stds->stds_len * 4;

 if ( (ret = uniraum_reqnewempty()) == NULL ) {
  err = roar_error;
  roar_stds_free(stds);
  roar_err_set(err);
  return NULL;
 }

 if ( roar_caps_to_msg(&(ret->mes), &caps, &(ret->long_data)) == -1 ) {
  err = roar_error;
  roar_stds_free(stds);
  uniraum_requnref(ret);
  roar_err_set(err);
  return NULL;
 }

 ret->mes.cmd = ROAR_CMD_CAPS;

 roar_stds_free(stds);

 return ret;
}

int              uniraum_reqget_type(uniraum_req_t * req) {
 _ckreq(req);
 return req->mes.cmd;
}

int              uniraum_reqget_stream(uniraum_req_t * req) {
 _ckreq(req);
 return req->mes.stream;
}

ssize_t          uniraum_reqget_datalen(uniraum_req_t * req) {
 _ckreq(req);
 return req->mes.datalen;
}

uniraum_pos_t    uniraum_reqget_pos(uniraum_req_t * req) {
 if ( req == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return 0;
 }

 return req->mes.pos;
}

void *           uniraum_reqget_data(uniraum_req_t * req) {
 if ( req == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return NULL;
 }

 if ( req->mes.cmd != ROAR_CMD_ADD_DATA ) {
  roar_error = ROAR_ERROR_INVAL;
  return NULL;
 }

 if ( req->long_data == NULL ) {
  return req->mes.data;
 } else {
  return req->long_data;
 }
}

uniraum_cap_t *  uniraum_reqget_caps(uniraum_req_t * req, size_t * len) {
 struct roar_caps caps;

 if ( req == NULL || len == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 if ( req->mes.cmd != ROAR_CMD_CAPS ) {
  roar_error = ROAR_ERROR_INVAL;
  return NULL;
 }

 if ( roar_caps_from_msg(&caps, &(req->mes), req->long_data) == -1 ) {
  return NULL;
 }

 if ( caps.version != 0 ) {
  roar_error = ROAR_ERROR_NSVERSION;
  return NULL;
 }

 if ( caps.type != ROAR_CT_STANDARDS ) {
  roar_error = ROAR_ERROR_BADRQC;
  return NULL;
 }

 if ( caps.flags != 0 ) {
  roar_error = ROAR_ERROR_NOTSUP;
  return NULL;
 }

 *len = caps.len / 4;

 return caps.data;
}

int              uniraum_reqget_stream_info(uniraum_req_t * req, int * parent, int * dir, struct roar_audio_info * info) {
 struct roar_stream sm;

 _ckreq(req);

 if ( parent == NULL || dir == NULL || info == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( req->mes.cmd != ROAR_CMD_NEW_STREAM ) {
  roar_error = ROAR_ERROR_TYPEMM;
  return -1;
 }

 memset(&sm, 0, sizeof(sm));

 if ( roar_stream_m2s(&sm, &(req->mes)) == -1 )
  return -1;

 *parent = sm.pos_rel_id;
 *dir    = sm.dir;

 memcpy(info, &(sm.info), sizeof(struct roar_audio_info));

 return 0;
}

int              uniraum_reqget_meta(uniraum_req_t * req, const char ** key, const char ** value, int * mode) {
 char * d;

 _ckreq(req);

 if ( key == NULL || value == NULL || mode == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 if ( req->mes.cmd != ROAR_CMD_SET_META ) {
  roar_error = ROAR_ERROR_TYPEMM;
  return -1;
 }

 if ( req->mes.datalen < 5 ) {
  roar_error = ROAR_ERROR_BADMSG;
  return -1;
 }

 if ( req->long_data == NULL ) {
  d = req->mes.data;
 } else {
  d = req->long_data;
 }

 if ( d[0] != 0 || d[3] != 0 ) {
  roar_error = ROAR_ERROR_BADMSG;
  return -1;
 }

 *mode = d[1];

 *key = roar_meta_strtype(d[2]);

 if ( d[4] == 0 ) {
  *value = NULL;
 } else {
  req->buf = roar_mm_malloc(d[4] + 1);

  if ( req->buf == NULL ) {
   return -1;
  }

  memcpy(req->buf, &(d[5+d[3]]), d[4]);
  ((char*)req->buf)[(int)d[4]] = 0;

  *value = req->buf;
 }

 return 0;
}

//ll
