//seektable.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "private.h"

#define _cktable()  if ( table == NULL ) { roar_err_set(ROAR_ERROR_FAULT); return -1; }

uniraum_seektable_t * uniraum_seektable_new(int stream, size_t points) {
 uniraum_seektable_t * ret = roar_mm_malloc(sizeof(uniraum_seektable_t));
 int err;

 if ( ret == NULL )
  return NULL;

 memset(ret, 0, sizeof(uniraum_seektable_t));

 ret->refc = 1;

 if ( points == 0 ) {
  ret->point = NULL;
 } else {
  ret->point = roar_mm_malloc(points*sizeof(uniraum_seekpoint_t*));
  if ( ret->point == NULL ) {
   err = roar_error;
   roar_mm_free(ret);
   roar_err_set(err);
   return NULL;
  }

  memset(ret->point, 0, points*sizeof(uniraum_seekpoint_t*));
 }

 ret->points = points;
 ret->stream = stream;

 return ret;
}

static int _read_bits(unsigned char * buf, int bits, uint64_t * val) {
 register uint64_t   ret    = 0;
 register uint64_t    c;

 if ( bits % 8 )
  return -1;

 for (; bits; ) {
  c = *(buf++);
//  fprintf(stderr, "data=%p, c=0x%.2X, ret=0x%.16llX, bits=%i\n", data, c, ret, bits);
  ret |= c << (bits -= 8);
 }

 *val = ret;
 return 0;
}

uniraum_seektable_t * uniraum_seektable_from_req(uniraum_req_t * req, uniraum_seekpos_t pos) {
 uniraum_seektable_t * ret = NULL;
 uniraum_seekpoint_t * point;
 uniraum_seekpos_t incrpos = pos;
 uniraum_pos_t reqpos;
 int reqstream;
 size_t reqlen;
 unsigned char * reqdata;
 size_t numpoints;
 size_t i;
 size_t cpos = 3;
 int err;
 int pevent, pflags, pstream;
 size_t plen, dlen;
 uint64_t tmp;
 int incr;

 if ( req == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 if ( uniraum_reqget_type(req) != ROAR_CMD_RAUM_SEEKTABLE ) {
  roar_err_set(ROAR_ERROR_TYPEMM);
  return NULL;
 }

 reqstream = uniraum_reqget_stream(req);
 reqlen    = uniraum_reqget_datalen(req);
 reqdata   = req->long_data == NULL ? req->mes.data : req->long_data;
 reqpos    = uniraum_reqget_pos(req);

 if ( reqlen < 3 ) {
  roar_err_set(ROAR_ERROR_BADCKSUM);
  return NULL;
 }

 if ( reqdata[0] != 0 ) {
  roar_err_set(ROAR_ERROR_NSVERSION);
  return NULL;
 }

 numpoints = (uint16_t)reqdata[1] << 8 | (uint16_t)reqdata[2];

 ret = uniraum_seektable_new(reqstream, numpoints);
 if ( ret == NULL )
  return NULL;

 ret->req_size = reqlen;

 for (i = 0; i < numpoints; i++) {
  if ( (reqlen - cpos) < 2 ) {
   uniraum_seektable_unref(ret);
   roar_err_set(ROAR_ERROR_BADCKSUM);
   return NULL;
  }

  pevent = reqdata[cpos++];
  pflags = reqdata[cpos++];

  if ( !(pflags & UNIRAUM_STFLAG_CMD) ) {
   pevent = -pevent - 1;
  }

  pstream = -1;
  if ( pflags & UNIRAUM_STFLAG_STREAMID ) {
   if ( (reqlen - cpos) < 2 ) {
    uniraum_seektable_unref(ret);
    roar_err_set(ROAR_ERROR_BADCKSUM);
    return NULL;
   }
   pstream = ((uint16_t)reqdata[cpos] << 8) | (uint16_t)reqdata[cpos+1];
   cpos += 2;
   if ( pstream == (uint16_t)-1 )
    pstream = -1;
  } else if ( reqstream != -1 ) {
   pstream = reqstream;
  }

  switch (pflags & UNIRAUM_STFLAG_SIZEMASK) {
   case UNIRAUM_STFLAG_SIZE16INC: plen =  2; incr = 1; break;
   case UNIRAUM_STFLAG_SIZE32INC: plen =  4; incr = 1; break;
   case UNIRAUM_STFLAG_SIZE32:    plen =  4; incr = 0; break;
   case UNIRAUM_STFLAG_SIZE64:    plen =  8; incr = 0; break;
   default:
     uniraum_seektable_unref(ret);
     roar_err_set(ROAR_ERROR_BADCKSUM);
     return NULL;
    break;
  }

  dlen = 2*(pflags & UNIRAUM_STFLAG_END ? 2 : 1)*plen;
  if ( (reqlen - cpos) < dlen ) {
   uniraum_seektable_unref(ret);
   roar_err_set(ROAR_ERROR_BADCKSUM);
   return NULL;
  }

  point = uniraum_seekpoint_new(pstream, pevent);
  if ( point == NULL ) {
   err = roar_error;
   uniraum_seektable_unref(ret);
   roar_err_set(err);
   return NULL;
  }

  _read_bits(&(reqdata[cpos+0*plen]), 8*plen, &tmp);
  point->pos_start = tmp;
  _read_bits(&(reqdata[cpos+1*plen]), 8*plen, &tmp);
  point->offset_start = tmp + pos;

  if ( incr) {
   point->pos_start    += reqpos;
   point->offset_start += incrpos;
   reqpos     = point->pos_start;
   incrpos    = point->offset_start;
  }

  if ( pflags & UNIRAUM_STFLAG_END ) {
   _read_bits(&(reqdata[cpos+2*plen]), 8*plen, &tmp);
   point->pos_end = tmp;
   _read_bits(&(reqdata[cpos+3*plen]), 8*plen, &tmp);
   point->offset_end = tmp;
   point->flags |= UNIRAUM_STFLAG_END;
   if ( incr) {
    point->pos_end    += reqpos;
    point->offset_end += incrpos;
   }
  }

  cpos += dlen;

  if ( uniraum_seektable_set_point(ret, i, point) == -1 ) {
   err = roar_error;
   uniraum_seekpoint_unref(point);
   uniraum_seektable_unref(ret);
   roar_err_set(err);
   return NULL;
  }

  uniraum_seekpoint_unref(point);
 }

 return ret;
}

uniraum_req_t       * uniraum_seektable_to_req(uniraum_seektable_t * table, uniraum_seekpos_t pos) {
 (void)table, (void)pos;
 roar_err_set(ROAR_ERROR_NOSYS);
 return NULL;
}


int                uniraum_seektable_ref(uniraum_seektable_t * table) {
 _cktable();

 table->refc++;

 return 0;
}

int                uniraum_seektable_unref(uniraum_seektable_t * table) {
 size_t i;

 _cktable();

 table->refc--;

 if ( table->refc == 0 ) {
  for (i = 0; i < table->points; i++)
   uniraum_seekpoint_unref(table->point[i]);

  if ( table->point != NULL )
   roar_mm_free(table->point);

  roar_mm_free(table);
 }

 return 0;
}

int                uniraum_seektable_set_stream(uniraum_seektable_t * table, int stream) {
 _cktable();

 table->stream = stream;

 return 0;
}

int                uniraum_seektable_get_stream(uniraum_seektable_t * table) {
 _cktable();

 return table->stream;
}

ssize_t               uniraum_seektable_num_points(uniraum_seektable_t * table) {
 _cktable();

 return table->points;
}

uniraum_seekpoint_t * uniraum_seektable_get_point(uniraum_seektable_t * table, size_t i) {
 if ( table == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 if ( i >= table->points ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 if ( uniraum_seekpoint_ref(table->point[i]) == -1 )
  return NULL;
 
 return table->point[i];
}

int                   uniraum_seektable_set_point(uniraum_seektable_t * table, size_t i, uniraum_seekpoint_t * point) {
 _cktable();

 if ( point == NULL || i >= table->points ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( table->point[i] != NULL )
  if ( uniraum_seekpoint_unref(table->point[i]) == -1 )
   return -1;

 table->point[i] = point;

 if ( uniraum_seekpoint_ref(table->point[i]) == -1 ) {
  table->point[i] = NULL;
  return -1;
 }

 return 0;
}

int                   uniraum_seektable_add_point(uniraum_seektable_t * table, uniraum_seekpoint_t * point) {
 size_t i;

 _cktable();

 if ( point == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 for (i = 0; i < table->points; i++) {
  if ( table->point[i] == NULL ) {
   if ( uniraum_seekpoint_ref(point) == -1 )
    return -1;

   table->point[i] = point;
   return 0;
  }
 }

 roar_err_set(ROAR_ERROR_NOMEM);
 return -1;
}

//ll
