//stream.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "private.h"

#define F_EXECED    0x0001
#define F_EOS       0x0002

#define _ckstr(s) if ( (s) == NULL ) { roar_error = ROAR_ERROR_FAULT; return -1; }

uniraum_stream_t * uniraum_stream_new(uniraum_req_t * req) {
 uniraum_stream_t * stream = roar_mm_malloc(sizeof(uniraum_stream_t));

 memset(stream, 0, sizeof(uniraum_stream_t));

 stream->refc   =  1;
 stream->flags  =  0;
 stream->stream = -1;
 stream->dir    = -1;
 stream->parent = -1;

 stream->req_queue_fill = 0;

 if ( req != NULL ) {
  if ( uniraum_stream_reqin(stream, req) == -1 ) {
   roar_mm_free(stream);
   return NULL;
  }
 }

 return stream;
}

int                uniraum_stream_ref(uniraum_stream_t * stream) {
 _ckstr(stream)

 stream->refc++;

 return 0;
}

int                uniraum_stream_unref(uniraum_stream_t * stream) {
 size_t i;

 _ckstr(stream);

 stream->refc--;

 if ( stream->refc == 0 ) {
  for (i = 0; i < stream->req_queue_fill; i++)
   uniraum_requnref(stream->req_queue[i]);
  roar_mm_free(stream);
 }

 return 0;
}

int                uniraum_stream_reqin(uniraum_stream_t * stream, uniraum_req_t * req) {
 int sid;
 int type;
 int error = 0;

 _ckstr(stream)

 if ( req == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return -1;
 }

 sid = uniraum_reqget_stream(req);
 type = uniraum_reqget_type(req);

 if ( sid == -1 )
  return -1;

 if ( stream->flags & F_EXECED ) {
  // if we are execed only accept ADD_DATA and KICK (EOS) packets which may be generated
  // by input rutines of application for more easy code.
  if ( type != ROAR_CMD_ADD_DATA && type != ROAR_CMD_KICK ) {
   roar_error = ROAR_ERROR_INVAL;
   return -1;
  }
 }

 if ( stream->req_queue_fill == _REQ_QUEUE_LEN ) {
  roar_error = ROAR_ERROR_AGAIN;
  return -1;
 }

 if ( stream->stream == -1 ) {
  if ( type != ROAR_CMD_NEW_STREAM ) {
   roar_error = ROAR_ERROR_TYPEMM;
   return -1;
  }
 } else {
  if ( stream->stream != sid ) {
   roar_error = ROAR_ERROR_INVAL;
   return -1;
  }
 }

 if ( uniraum_reqref(req) == -1 )
  return -1;

 switch (type) {
  case ROAR_CMD_NEW_STREAM:
    if ( uniraum_reqget_stream_info(req, &(stream->parent), &(stream->dir), &(stream->info)) == -1 ) {
     error = 1;
    }
   break;
  case ROAR_CMD_EXEC_STREAM:
    stream->flags |= F_EXECED;
   break;
  case ROAR_CMD_KICK:
    stream->flags |= F_EOS;
   break;
 }

 if ( error ) {
  error = roar_error;
  uniraum_requnref(req);
  roar_error = error;
  return -1;
 }

 if ( stream->stream == -1 ) {
  stream->stream = sid;
 }

 stream->req_queue[stream->req_queue_fill++] = req;

 return 0;
}

uniraum_req_t *    uniraum_stream_reqout(uniraum_stream_t * stream) {
 uniraum_req_t * req;
 size_t i;

 if ( stream == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return NULL;
 }

 if ( stream->req_queue_fill == 0 ) {
  roar_error = ROAR_ERROR_AGAIN;
  return NULL;
 }

 req = stream->req_queue[0];

 stream->req_queue_fill--;

 for (i = 0; i < stream->req_queue_fill; i++)
  stream->req_queue[i] = stream->req_queue[i+1];

 return req;
}


int                uniraum_stream_get_stream(uniraum_stream_t * stream) {
 _ckstr(stream);
 if ( stream->stream == -1 )
  roar_error = ROAR_ERROR_INVAL;
 return stream->stream;
}

int                uniraum_stream_get_parent(uniraum_stream_t * stream) {
 _ckstr(stream);
 if ( stream->stream == -1 )
  roar_error = ROAR_ERROR_INVAL;
 return stream->parent;
}

int                uniraum_stream_get_dir(uniraum_stream_t * stream) {
 _ckstr(stream);
 if ( stream->stream == -1 )
  roar_error = ROAR_ERROR_INVAL;
 return stream->dir;
}

const struct roar_audio_info * uniraum_stream_get_auinfo(uniraum_stream_t * stream) {
 if ( stream == NULL ) {
  roar_error = ROAR_ERROR_FAULT;
  return NULL;
 }

 if ( stream->stream == -1 ) {
  roar_error = ROAR_ERROR_INVAL;
  return NULL;
 }

 return &(stream->info);
}

int                uniraum_stream_is_execed(uniraum_stream_t * stream) {
 _ckstr(stream);

 if ( stream->stream == -1 ) {
  roar_error = ROAR_ERROR_INVAL;
  return -1;
 }

 return stream->flags & F_EXECED ? 1 : 0;
}

int                uniraum_stream_is_eos(uniraum_stream_t * stream) {
 _ckstr(stream);

 if ( stream->stream == -1 ) {
  roar_error = ROAR_ERROR_INVAL;
  return -1;
 }

 return stream->flags & F_EOS ? 1 : 0;
}

ssize_t            uniraum_stream_pending(uniraum_stream_t * stream) {
 _ckstr(stream);

 return stream->req_queue_fill;
}

//ll
