//uniraum-enc.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include <uniraum/uniraum.h>

#define _AS(x) (sizeof((x))/sizeof(*(x)))

#define _CKHAVEARGS(x) if ( (i + (x)) >= argc ) { \
                        ROAR_ERR("Option %s requires more arguments. See --help for more details.", k); \
                        return 70; \
                       }

struct meta {
 char * key;
 char * value;
};

#define META_MAX  64

#define ABUFS_MAX (2*META_MAX)
static char * g_abufs[ABUFS_MAX];
static size_t g_abufp = 0;

static char * _astrdup(const char * s) {
 char * ret;

 if ( g_abufp == ABUFS_MAX ) {
  roar_err_set(ROAR_ERROR_NOMEM);
  return NULL;
 }

 ret = roar_mm_strdup(s);
 if ( ret == NULL )
  return NULL;

 g_abufs[g_abufp++] = ret;

 return ret;
}

static void _afree(void) {
 size_t i;

 for (i = 0; i < g_abufp; i++)
  roar_mm_free(g_abufs[i]);

 g_abufp = 0;
}

static void usage (const char * progname) {
 fprintf(stderr, "Usage: %s [OPTIONS] {-o|--output} outputfile.raum inputfile\n", progname);
 fprintf(stderr, "\nOptions:\n\n");

 fprintf(stderr,
         " -h --help                       - Show this help\n"
         " -o --output FILE                - Set output file\n"
         "\n"
         " -R RATE                         - Set sample rate of input\n"
         " -B BITS                         - Set bits per sample of input\n"
         " -C CHANNELS                     - Set channels per frame of input\n"
         " -E CODEC                        - Set codec of input\n"
         "    --aiprofile PROFILE          - Use Audio Info profile\n"
         "\n"
         "    --padding PADDING            - Set size of global and stream padding (in bytes)\n"
         "    --padding-global PADDING     - Set size of global padding (in bytes)\n"
         "    --padding-stream PADDING     - Set size of stream padding (in bytes)\n"
         "\n"
         "    --caps                       - Adds a capabilities req (default)\n"
         "    --no-caps                    - Do not add a capabilities req\n"
         "\n"
         "    --comment-global KEY=VAL     - Add a global comment (meta data)\n"
         "    --import-comment-global FILE - Add global comments from file\n"
         "    --comment-stream KEY=VAL     - Add a stream specific comment (meta data)\n"
         "    --import-comment-stream FILE - Add stream specific comments from file\n"
        );
}

static int run(uniraum_file_t * out, struct roar_vio_calls * vio, struct roar_audio_info * info, int dir, size_t padding, struct meta * meta, size_t metalen) {
 uniraum_mapping_t * mapping;
 uniraum_req_t     * req;
 uniraum_data_t    * data;
 uniraum_pos_t       pos    = 0;
 int                 stream = 0;
 size_t i;
 ssize_t ret;
 char buf[1200];
 void * data_data;

 mapping = uniraum_mapping_new(stream, UNIRAUM_OUT, ROAR_VIO_DFT_RAW, info);
 if ( mapping == NULL )
  return -1;

 req = uniraum_reqnew_stream(stream, -1, dir, info);
 uniraum_reqout(out, req);
 uniraum_requnref(req);

 for (i = 0; i < metalen; i++) {
  req = uniraum_reqadd_meta(stream, meta[i].key, meta[i].value);
  uniraum_reqout(out, req);
  uniraum_requnref(req);
 }

 req = uniraum_reqfin_meta(stream);
 uniraum_reqout(out, req);
 uniraum_requnref(req);

 if ( padding ) {
  req = uniraum_reqnoop(padding, stream, 0, 0);
  uniraum_reqout(out, req);
  uniraum_requnref(req);
 }

 while ((ret = roar_vio_read(vio, buf, sizeof(buf))) > 0) {
  data = uniraum_data_new(ret, &data_data, ROAR_VIO_DFT_RAW);
  memcpy(data_data, buf, ret);

  uniraum_mapping_datain(mapping, data);
  uniraum_data_unref(data);

  while ((req = uniraum_mapping_reqout(mapping)) != NULL) {
   uniraum_reqout(out, req);
   uniraum_requnref(req);
  }
 }

 pos = uniraum_mapping_get_pos(mapping);

 // write out a empty request for compatibily with older software.
 req = uniraum_reqdata(stream, pos, buf, 0);
 uniraum_reqout(out, req);
 uniraum_requnref(req);

 req = uniraum_reqend_stream(stream, pos);
 uniraum_reqout(out, req);
 uniraum_requnref(req);

 uniraum_mapping_unref(mapping);

 return 0;
}

static int read_meta_from_file(const char * file, struct meta * meta, size_t * metanum) {
 struct roar_vio_defaults def;
 struct roar_vio_calls vio;
 char line[2048];
 size_t len;

 if ( roar_vio_dstr_init_defaults(&def, ROAR_VIO_DEF_TYPE_NONE, O_RDONLY, 0644) == -1 ) {
  ROAR_ERR("Can not init DSTR defaults. Bad.");
  return -1;
 }

 if ( roar_vio_open_dstr(&vio, (char*)file, &def, 1) == -1 ) {
  ROAR_ERR("Can not open meta data file: %s: %s", file, roar_error2str(roar_error));
  return -1;
 }

 while (roar_vio_fgets(&vio, line, sizeof(line)) != NULL) {
  line[sizeof(line)-1] = 0;

  len = roar_mm_strlen(line);
  for (len--; len && (line[len] == '\r' || line[len] == '\n'); len--)
   line[len] = 0;

  if ( *metanum == META_MAX ) {
   ROAR_ERR("Too many meta data entry given.");
   break;
  }
  meta[*metanum].key   = _astrdup(line);

  if ( meta[*metanum].key == NULL ) {
   ROAR_ERR("Can not allocate meta data buffer: %s", roar_error2str(roar_error));
   break;
  }

  meta[*metanum].value = strstr(meta[*metanum].key, "=");

  if ( meta[*metanum].value == NULL ) {
   ROAR_ERR("Meta data format error.");
   continue;
  }

  *meta[*metanum].value = 0;
  meta[*metanum].value++;

  (*metanum)++;
 }

 roar_vio_close(&vio);

 return 0;
}

int main (int argc, char * argv[]) {
 struct roar_vio_defaults def;
 struct roar_vio_calls in_vio;
 struct roar_audio_info info;
 uniraum_file_t * out;
 uniraum_req_t  * req;
 int codec = -1;
 int dir = ROAR_DIR_PLAY;
 const char * infile  = NULL;
 const char * outfile = NULL;
 size_t padding_global = 512;
 size_t padding_stream = 512;
 struct meta meta_global[META_MAX], meta_stream[META_MAX];
 size_t meta_global_num = 0, meta_stream_num = 0;
 int caps = 1;
 char * k;
 int i;
 size_t is;
 ssize_t ret;
 char buf[64];

 atexit(_afree);
 
 memset(&info, 0, sizeof(info));

 for (i = 1; i < argc; i++) {
  k = argv[i];
  if ( !strcmp(k, "-h") || !strcmp(k, "--help") ) {
   usage(argv[0]);
   return 0;
  } else if ( !strcmp(k, "-o") || !strcmp(k, "--output") ) {
   _CKHAVEARGS(1);
   outfile = argv[++i];

  } else if ( !strcmp(k, "-R") ) {
   _CKHAVEARGS(1);
   info.rate = roar_str2rate(argv[++i]);
  } else if ( !strcmp(k, "-B") ) {
   _CKHAVEARGS(1);
   info.bits = roar_str2bits(argv[++i]);
  } else if ( !strcmp(k, "-C") ) {
   _CKHAVEARGS(1);
   info.channels = roar_str2channels(argv[++i]);
  } else if ( !strcmp(k, "-E") ) {
   _CKHAVEARGS(1);
   codec = roar_str2codec(argv[++i]);

  } else if ( !strcmp(k, "--padding") ) {
   _CKHAVEARGS(1);
   padding_global = atoi(argv[++i]);
   padding_stream = atoi(argv[i]);
  } else if ( !strcmp(k, "--padding-global") ) {
   _CKHAVEARGS(1);
   padding_global = atoi(argv[++i]);
  } else if ( !strcmp(k, "--padding-stream") ) {
   _CKHAVEARGS(1);
   padding_stream = atoi(argv[++i]);

  } else if ( !strcmp(k, "--caps") ) {
   caps = 1;
  } else if ( !strcmp(k, "--no-caps") ) {
   caps = 0;

  } else if ( !strcmp(k, "--aiprofile") ) {
   _CKHAVEARGS(1);
   if ( roar_profile2info(&info, argv[++i]) == -1 ) {
    ROAR_ERR("Audio profile %s not known.", argv[i]);
    return 2;
   }
   codec = info.codec;

  } else if ( !strcmp(k, "--comment-global") ) {
   _CKHAVEARGS(1);
   if ( meta_global_num == META_MAX ) {
    ROAR_ERR("Too many meta data entry given.");
    return 2;
   }
   meta_global[meta_global_num].key   = argv[++i];
   meta_global[meta_global_num].value = strstr(argv[i], "=");

   if ( meta_global[meta_global_num].value == NULL ) {
    ROAR_ERR("Meta data format error.");
    return 2;
   }

   *meta_global[meta_global_num].value = 0;
   meta_global[meta_global_num].value++;

   meta_global_num++;
  } else if ( !strcmp(k, "--import-comment-global") ) {
   _CKHAVEARGS(1);
   read_meta_from_file(argv[++i], meta_global, &meta_global_num);

  } else if ( !strcmp(k, "--comment-stream") ) {
   _CKHAVEARGS(1);
   if ( meta_stream_num == META_MAX ) {
    ROAR_ERR("Too many meta data entry given.");
    return 2;
   }
   meta_stream[meta_stream_num].key   = argv[++i];
   meta_stream[meta_stream_num].value = strstr(argv[i], "=");

   if ( meta_stream[meta_stream_num].value == NULL ) {
    ROAR_ERR("Meta data format error.");
    return 2;
   }

   *meta_stream[meta_stream_num].value = 0;
   meta_stream[meta_stream_num].value++;

   meta_stream_num++;
  } else if ( !strcmp(k, "--import-comment-stream") ) {
   _CKHAVEARGS(1);
   read_meta_from_file(argv[++i], meta_stream, &meta_stream_num);

  } else if ( infile == NULL ) {
   infile = k;
  } else {
   usage(argv[0]);
   return 1;
  }
 }

 if ( roar_vio_dstr_init_defaults(&def, ROAR_VIO_DEF_TYPE_NONE, O_RDONLY, 0644) == -1 ) {
  ROAR_ERR("Can not init DSTR defaults. Bad.");
  return 3;
 }

 if ( roar_vio_open_dstr(&in_vio, (char*)infile, &def, 1) == -1 ) {
  ROAR_ERR("Can not open file: %s: %s", infile, roar_error2str(roar_error));
  return 3;
 }

 if ( codec == -1 ) {
  roar_err_set(ROAR_ERROR_NONE);
  ret = roar_vio_read(&in_vio, buf, sizeof(buf));

  if (ret < 1) {
   ROAR_ERR("Can not read data from input stream: %s", roar_error2str(roar_error));
   roar_vio_close(&in_vio);
   return 3;
  }

  codec = roar_file_codecdetect(buf, ret);

  if ( roar_vio_lseek(&in_vio, 0, SEEK_SET) != 0 ) {
   ROAR_ERR("Can not seek in input stream for codec detection: %s", roar_error2str(roar_error));
   roar_vio_close(&in_vio);
   return 3;
  }

  if ( codec == -1 ) {
   ROAR_ERR("Can not detect codec. Try setting it using -E");
   roar_vio_close(&in_vio);
   return 3;
  }

  switch (codec) {
   case ROAR_CODEC_OGG_VORBIS: codec = ROAR_CODEC_RAUM_VORBIS; break;
   case ROAR_CODEC_OGG_FLAC:   codec = ROAR_CODEC_RAUM_FLAC;   break;
  }
 }

 info.codec = codec;

 if ( info.rate == 0 ) {
  switch (info.codec) {
   case ROAR_CODEC_PCM_S_LE:
   case ROAR_CODEC_PCM_S_BE:
   case ROAR_CODEC_PCM_S_PDP:
   case ROAR_CODEC_PCM_U_LE:
   case ROAR_CODEC_PCM_U_BE:
   case ROAR_CODEC_PCM_U_PDP:
   case ROAR_CODEC_AUTLAW_LE:
   case ROAR_CODEC_AUTLAW_BE:
   case ROAR_CODEC_MUUTLAW_LE:
   case ROAR_CODEC_MUUTLAW_BE:
     info.rate = ROAR_RATE_DEFAULT;
    break;
   case ROAR_CODEC_ALAW:
   case ROAR_CODEC_MULAW:
     info.bits = 8000;
    break;
  }
 }

 if ( info.bits == 0 ) {
  switch (info.codec) {
   case ROAR_CODEC_PCM_S_LE:
   case ROAR_CODEC_PCM_S_BE:
   case ROAR_CODEC_PCM_S_PDP:
   case ROAR_CODEC_PCM_U_LE:
   case ROAR_CODEC_PCM_U_BE:
   case ROAR_CODEC_PCM_U_PDP:
   case ROAR_CODEC_AUTLAW_LE:
   case ROAR_CODEC_AUTLAW_BE:
   case ROAR_CODEC_MUUTLAW_LE:
   case ROAR_CODEC_MUUTLAW_BE:
     info.bits = 16;
    break;
   case ROAR_CODEC_ALAW:
   case ROAR_CODEC_MULAW:
     info.bits = 8;
    break;
   case ROAR_CODEC_RAUM_VORBIS:
     info.bits = 32;
    break;
  }
 }

 if ( info.channels == 0 ) {
  switch (info.codec) {
   case ROAR_CODEC_PCM_S_LE:
   case ROAR_CODEC_PCM_S_BE:
   case ROAR_CODEC_PCM_S_PDP:
   case ROAR_CODEC_PCM_U_LE:
   case ROAR_CODEC_PCM_U_BE:
   case ROAR_CODEC_PCM_U_PDP:
   case ROAR_CODEC_AUTLAW_LE:
   case ROAR_CODEC_AUTLAW_BE:
   case ROAR_CODEC_MUUTLAW_LE:
   case ROAR_CODEC_MUUTLAW_BE:
     info.channels = ROAR_CHANNELS_DEFAULT;
    break;
   case ROAR_CODEC_ALAW:
   case ROAR_CODEC_MULAW:
     info.channels = 1;
    break;
  }
 }

 if ( info.rate == 0 ) {
   ROAR_ERR("Input sample rate not set, Set it with -R");
  roar_vio_close(&in_vio);
  return 4;
 }

 if ( info.bits == 0 ) {
   ROAR_ERR("Input sample size not set, Set it with -B");
  roar_vio_close(&in_vio);
  return 4;
 }

 if ( info.channels == 0 ) {
   ROAR_ERR("Input's number of channels not set, Set it with -C");
  roar_vio_close(&in_vio);
  return 4;
 }

 out = uniraum_opendstr(outfile, NULL, O_WRONLY|O_CREAT);
 if ( out == NULL ) {
  ROAR_ERR("Can not open output file: %s: %s", outfile, roar_error2str(uniraum_error(NULL)));
  roar_vio_close(&in_vio);
  return 3;
 }

 if ( caps ) {
  req = uniraum_reqcaps(UNIRAUM_CAPS_MINIMUM, NULL, 0);
  uniraum_reqout(out, req);
  uniraum_requnref(req);
 }

 for (is = 0; is < meta_global_num; is++) {
  req = uniraum_reqadd_meta(-1, meta_global[is].key, meta_global[is].value);
  uniraum_reqout(out, req);
  uniraum_requnref(req);
 }

 req = uniraum_reqfin_meta(-1);
 uniraum_reqout(out, req);
 uniraum_requnref(req);

 if ( padding_global ) {
  req = uniraum_reqpadding(padding_global);
  uniraum_reqout(out, req);
  uniraum_requnref(req);
 }

 run(out, &in_vio, &info, dir, padding_stream, meta_stream, meta_stream_num);

 uniraum_close(out);

 roar_vio_close(&in_vio);

 return 0;
}

//ll
