//uniraum-info.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include <uniraum/uniraum.h>

int g_verbose = 0;

static void usage(char * progname) {
 fprintf(stderr, "Usage: %s [OPTIONS] file [file [file [...]]]\n", progname);
 fprintf(stderr, "\nOptions:\n");
 fprintf(stderr,
                 " -h --help             - Print this help.\n"
                 " -v --verbose          - Be verbose. Can be used multiple times.\n"
        );
}

#define F_STREAM   0x0001
#define F_DATA     0x0002
#define F_POS      0x0004

#define F_PARSE    0x0100
#define F_NOTOPEN  0x0200

static const char * _caps2str(uniraum_cap_t * caps, size_t len) {
 static char buf[1024] = "";
 char capbuf[80];
 uniraum_cap_t ho;
 unsigned int vendor, standard, version;
 const char * vendor_name;
 size_t i;
 size_t capbuflen;
 size_t fill = 0;

 for (i = 0; i < len; i++) {
  ho = ROAR_NET2HOST32(caps[i]);

  vendor   = ROAR_STD_VENDOR(ho);
  standard = ROAR_STD_STD(ho);
  version  = ROAR_STD_VERSION(ho);

  vendor_name = uniraum_cap_vendor2str(vendor);
  if ( vendor_name != NULL ) {
   snprintf(capbuf, sizeof(capbuf), "%s-%u-%u, ", vendor_name, standard, version);
  } else {
   snprintf(capbuf, sizeof(capbuf), "%u-%u-%u, ", vendor, standard, version);
  }

  capbuf[sizeof(capbuf)-1] = 0;

  capbuflen = roar_mm_strlen(capbuf);

  if ( capbuflen > (sizeof(buf) - 1 - fill) ) {
   buf[sizeof(buf)-4] = '.';
   buf[sizeof(buf)-3] = '.';
   buf[sizeof(buf)-2] = '.';
   buf[sizeof(buf)-1] = 0;
   return buf;
  } else {
   memcpy(buf+fill, capbuf, capbuflen);
   fill += capbuflen;
  }
 }

 buf[fill-2] = 0;
 return buf;
}

#define MAX32 2147483647LL
#define llX (long long unsigned int)

static int print_seektable(size_t reqcount, uniraum_seektable_t * seektable) {
 uniraum_seekpoint_t * point;
 ssize_t i, numpoints;
 int ptype, pstream;
 char streambuf[32];
 const char * typename;
 uniraum_pos_t spos, epos;
 uniraum_seekpos_t sspos, espos;
 int has_end;
 int show_pos;
 char posbuf[128];

 numpoints = uniraum_seektable_num_points(seektable);

 for (i = 0; i < numpoints; i++) {
  point = uniraum_seektable_get_point(seektable, i);

  if ( point == NULL ) {
   printf("req %4zu (seekpoint): Point %zi: -- no data --\n", reqcount, i);
   continue;
  }

  ptype = uniraum_seekpoint_get_type(point);
  pstream = uniraum_seekpoint_get_stream(point);

  uniraum_seekpoint_get_startpos(point, &spos, &sspos);

  if ( uniraum_seekpoint_get_endpos(point, &epos, &espos) == -1 ) {
   has_end = 0;
  } else {
   has_end = 1;
  }

  switch (ptype) {
   case UNIRAUM_SEEKPOINT_STUB:   typename = "stub";            show_pos = 0; break;
   case UNIRAUM_SEEKPOINT_EXECED: typename = "execed data";     show_pos = 1; break;
   case ROAR_CMD_NOOP:            typename = "padding";         show_pos = 0; break;
   case ROAR_CMD_NEW_STREAM:      typename = "new stream";      show_pos = 0; break;
   case ROAR_CMD_SET_META:        typename = "meta data event"; show_pos = 1; break;
   case ROAR_CMD_EXEC_STREAM:     typename = "exec stream";     show_pos = 1; break;
   case ROAR_CMD_QUIT:            typename = "EOF Mark";        show_pos = 0; break;
   case ROAR_CMD_ADD_DATA:        typename = "data";            show_pos = 1; break;
   case ROAR_CMD_KICK:            typename = "end of stream";   show_pos = 1; break;
   case ROAR_CMD_CAPS:            typename = "caps";            show_pos = 0; break;
   case ROAR_CMD_RAUM_SEEKTABLE:  typename = "seektable";       show_pos = 1; break;
   default: typename = "<<unknown>>"; show_pos = 1; break;
  }

  if ( pstream == -1 ) {
   strncpy(streambuf, "no stream", sizeof(streambuf));
  } else {
   snprintf(streambuf, sizeof(streambuf), "stream %2i", pstream);
  }
  streambuf[sizeof(streambuf)-1] = 0;

  if ( ptype == UNIRAUM_SEEKPOINT_STUB ) {
   strncpy(posbuf, "no position data", sizeof(posbuf));
  } else if ( has_end ) {
   if ( spos < MAX32 && sspos < MAX32 && epos < MAX32 && espos < MAX32 ) {
    if ( show_pos ) {
     snprintf(posbuf, sizeof(posbuf), "at 0x%.8llX/0x%.8llX -> 0x%.8llX/0x%.8llX", llX sspos, llX spos, llX espos, llX epos);
    } else {
     snprintf(posbuf, sizeof(posbuf), "at 0x%.8llX -> 0x%.8llX", llX sspos, llX espos);
    }
   } else {
    snprintf(posbuf, sizeof(posbuf), "at 0x%.16llX/0x%.16llX -> 0x%.16llX/0x%.16llX", llX sspos, llX spos, llX espos, llX epos);
   }
  } else {
   if ( spos < MAX32 && sspos < MAX32 ) {
    if ( show_pos ) {
     snprintf(posbuf, sizeof(posbuf), "at 0x%.8llX/0x%.8llX", llX sspos, llX spos);
    } else {
     snprintf(posbuf, sizeof(posbuf), "at 0x%.8llX", llX sspos);
    }
   } else {
    snprintf(posbuf, sizeof(posbuf), "at 0x%.16llX/0x%.16llX", llX sspos, llX spos);
   }
  }
  posbuf[sizeof(posbuf)-1] = 0;

  printf("req %4zu (seekpoint, %s): Point %zi of type %s: %s\n", reqcount, streambuf, i, typename, posbuf);
 }

 return 0;
}

int do_file(char * file) {
 uniraum_file_t * inp;
 uniraum_req_t * req;
 size_t reqcount = 0;
 int type, stream;
 ssize_t datalen;
 const char * typename;
 uniraum_pos_t pos;
 int min_verbose;
 int f_need, f_no, f_have, f_res;
 int f_probs;
 int have_errors;
 char extra_text[1024];
 char at_text[32];
 uniraum_seekpos_t curpos, curpos_next;

 // stream store:
 struct stream {
  int id;
  struct roar_audio_info info;
 } streams[32], * cstream;
 size_t i;

 // for stats:
 size_t c_data = 0, c_data_nonempty = 0;
 size_t l_all = 0, l_data = 0;

 // for NEW_STREAM:
 struct roar_audio_info info;
 int parent;
 int dir;

 // for KICK:
 float    playback_length;
 int      pl_h, pl_m;

 // for SET_META:
 const char * key, * value;
 int mode;

 // for CAPS:
 uniraum_cap_t * caps;
 size_t capslen;

 // for EXEC_STREAM:
 int execed = -1;

 // for QUIT:
 int eof_mark = 0;

 // for SEEKTABLE:
 uniraum_seektable_t * seektable;


 memset(streams, 0, sizeof(streams));
 for (i = 0; i < (sizeof(streams)/sizeof(*streams)); i++)
  streams[i].id = -1;

 if ( file == NULL ) {
  inp = uniraum_openvio(roar_stdin, O_RDONLY);
 } else {
  inp = uniraum_opendstr(file, NULL, O_RDONLY);
 }

 if ( inp == NULL ) {
  fprintf(stderr, "Can not open input file: %s: %s\n", file, roar_error2str(uniraum_error(NULL)));
  return -1;
 }

 curpos_next  = uniraum_get_seekpos(inp);

 while (uniraum_read(inp) >= 0) {
  while ((req = uniraum_reqin(inp)) != NULL) {
   curpos       = curpos_next;
   curpos_next  = uniraum_get_seekpos(inp);

   type    = uniraum_reqget_type(req);
   stream  = uniraum_reqget_stream(req);
   datalen = uniraum_reqget_datalen(req);
   pos     = uniraum_reqget_pos(req);

   l_all  += datalen;

   if ( type == ROAR_CMD_ADD_DATA ) {
    l_data += datalen;
    c_data++;
    if ( datalen != 0 )
     c_data_nonempty++;
   }

   f_need = f_no = f_have = 0;
   f_probs = 0;
   have_errors = 0;

   extra_text[0] = 0;

   cstream = NULL;
   for (i = 0; cstream == NULL && i < (sizeof(streams)/sizeof(*streams)); i++)
    if ( streams[i].id == stream )
     cstream = &(streams[i]);

   if ( cstream == NULL && type != ROAR_CMD_NEW_STREAM ) {
    f_probs |= F_NOTOPEN;
   }

   for (i = 0; cstream == NULL && i < (sizeof(streams)/sizeof(*streams)); i++) {
    if ( streams[i].id == -1 ) {
     cstream = &(streams[i]);
     memset(cstream, 0, sizeof(*cstream));
     cstream->id = stream;
    }
   }

   if ( cstream == NULL ) {
    fprintf(stderr, "Too many streams open in parallel for this tool. Your file may be corrupted."
                    " Killing an old stream.\n");
    cstream = &(streams[0]);
    memset(cstream, 0, sizeof(*cstream));
    cstream->id = stream;
   }

   seektable = NULL;

   switch (type) {
    case ROAR_CMD_NOOP:
      typename = "padding";
      min_verbose = 1;
     break;
    case ROAR_CMD_QUIT:
      typename = "eof";
      min_verbose = 1;
      f_no = F_STREAM|F_DATA|F_POS;
      eof_mark = 1;
     break;
    case ROAR_CMD_SET_META:
      typename = "meta";
      min_verbose = 0;
      f_need = F_DATA;
      if ( uniraum_reqget_meta(req, &key, &value, &mode) == -1 ) {
       f_probs |= F_PARSE;
      } else {
       if ( mode == ROAR_META_MODE_SET || mode == ROAR_META_MODE_ADD ) {
        snprintf(extra_text, sizeof(extra_text), "%s meta data %s: %s",
                   mode == ROAR_META_MODE_SET ? "Setting" : "Adding",
                   key, value);
       } else if ( mode == ROAR_META_MODE_FINALIZE ) {
        snprintf(extra_text, sizeof(extra_text), "Meta data is finalized.");
        min_verbose = 2;
       } else if ( mode == ROAR_META_MODE_CLEAR ) {
        snprintf(extra_text, sizeof(extra_text), "Meta data is cleared.");
        min_verbose = 1;
       } else {
        f_probs |= F_PARSE;
       }
      }
     break;
    case ROAR_CMD_EXEC_STREAM:
      typename = "exec";
      min_verbose = 1;
      f_need = F_STREAM;
      f_no   = F_DATA;
      execed = stream;
     break;
    case ROAR_CMD_RAUM_SEEKTABLE:
      typename = "seektable";
      min_verbose = 1;
      f_need = F_DATA;
      if ( (seektable = uniraum_seektable_from_req(req, curpos)) == NULL ) {
       f_probs |= F_PARSE;
      } else {
       snprintf(extra_text, sizeof(extra_text), "Points: %zi", uniraum_seektable_num_points(seektable));
      }
     break;
    case ROAR_CMD_NEW_STREAM:
      typename = "new_stream";
      min_verbose = 0;
      f_need = F_STREAM|F_DATA;
      f_no   = F_POS;
      if ( uniraum_reqget_stream_info(req, &parent, &dir, &info) == -1 ) {
       f_probs |= F_PARSE;
      } else {
       memcpy(&(cstream->info), &info, sizeof(info));
       if ( parent == -1 ) {
        snprintf(extra_text, sizeof(extra_text), "%uch/%uBit/%uHz, codec=%u(%s), dir=%i(%s)",
                     info.channels, info.bits, info.rate,
                     info.codec, roar_codec2str(info.codec),
                     dir, roar_dir2str(dir));
       } else {
        snprintf(extra_text, sizeof(extra_text), "%uch/%uBit/%uHz, codec=%u(%s), dir=%i(%s), parent=%i",
                     info.channels, info.bits, info.rate,
                     info.codec, roar_codec2str(info.codec),
                     dir, roar_dir2str(dir),
                     parent);
       }
      }
     break;
    case ROAR_CMD_KICK:
      typename = "end_stream";
      min_verbose = 0;
      f_need = F_STREAM|F_DATA;
      playback_length  = (float)pos/cstream->info.rate;
      playback_length -= (pl_h = (int)(playback_length/3600))*3600;
      playback_length -= (pl_m = (int)(playback_length/  60))*  60;
      snprintf(extra_text, sizeof(extra_text), "playback length: %.2i:%.2i:%05.2f sec", pl_h, pl_m, playback_length);
      cstream->id = -1; // stream ends here.
     break;
    case ROAR_CMD_ADD_DATA:
      typename = "data";
      min_verbose = 5;
      f_need = F_STREAM;
      if ( g_verbose > min_verbose ) {
       playback_length  = (float)pos/cstream->info.rate;
       playback_length -= (pl_h = (int)(playback_length/3600))*3600;
       playback_length -= (pl_m = (int)(playback_length/  60))*  60;
      snprintf(extra_text, sizeof(extra_text), "playback position: %.2i:%.2i:%05.2f sec", pl_h, pl_m, playback_length);
      }
     break;
    case ROAR_CMD_CAPS:
      typename = "caps";
      min_verbose = 2;
      f_need = F_DATA;
      f_no   = F_STREAM|F_POS;
      if ( (caps = uniraum_reqget_caps(req, &capslen)) == NULL ) {
       f_probs |= F_PARSE;
      } else {
       snprintf(extra_text, sizeof(extra_text), "%llu Caps: %s", (long long unsigned int)capslen,
                                                                 _caps2str(caps, capslen));
      }
     break;
    default:
      typename = "(unknown)";
      min_verbose = 0;
     break;
   }

   if ( stream != -1 )
    f_have |= F_STREAM;
   if ( datalen > 0 )
    f_have |= F_DATA;
   if ( pos > 0 )
    f_have |= F_POS;

   if ( (f_have & f_no) || ((f_have | f_need) != f_have) || f_probs )
    have_errors = 1;

   uniraum_requnref(req);

   extra_text[sizeof(extra_text)-1] = 0;

   if ( g_verbose >= min_verbose || have_errors ) {
    if ( g_verbose ) {
     snprintf(at_text, sizeof(at_text), " at 0x%.6X", (unsigned int)curpos);
     at_text[sizeof(at_text)-1] = 0;
    } else {
     at_text[0] = 0;
    }
    if ( stream == -1 ) {
     printf("req %4zu%s (type %3i/%s, datalen %3zi Byte, no stream)%s%s\n", reqcount, at_text,
             type, typename, datalen, extra_text[0] ? ": " : "", extra_text);
    } else {
     if ( g_verbose >= 2 ) {
      printf("req %4zu%s (type %3i/%s, datalen %3zi Byte, stream %i, pos=%llu)%s%s\n", reqcount, at_text,
              type, typename, datalen, stream, (long long unsigned int)pos, extra_text[0] ? ": " : "", extra_text);
     } else {
      printf("req %4zu%s (type %3i/%s, datalen %3zi Byte, stream %i)%s%s\n", reqcount, at_text,
              type, typename, datalen, stream, extra_text[0] ? ": " : "", extra_text);
     }
    }

    // show seektable ponts if we have a seektable and verboseity is bigger than minimum for seektables.
    if ( seektable != NULL && g_verbose > min_verbose ) {
     print_seektable(reqcount, seektable);
    }

    if ( have_errors ) {
     printf(" \\--> Request has errors: ");

     if ( f_probs & F_PARSE )
      printf("Request can not be parsed. ");
     if ( f_probs & F_NOTOPEN )
      printf("Request uses a non open stream. ");

     f_res = f_have & f_no;
     if ( f_res & F_STREAM )
      printf("Has stream ID set but is not allowed to have. ");
     if ( f_res & F_DATA )
      printf("Has data but is not allowed to have. ");
     if ( f_res & F_POS )
      printf("Has position but is not allowed to have. ");

     f_res = (f_have | f_need) != f_have;
     if ( f_res & F_STREAM )
      printf("Does not set stream ID but needs to. ");
     if ( f_res & F_DATA )
      printf("Does not have data but needs to. ");
     if ( f_res & F_POS )
      printf("Does not have position but needs to. ");

     printf("\n");
    }
   }

   if ( seektable != NULL )
    uniraum_seektable_unref(seektable);

   reqcount++;
  }
 }

 uniraum_close(inp);

 if ( !eof_mark && execed == -1 ) {
  printf("Error: No EOF Mark was set nor any stream execed. File may be corrupted.\n");
 }

 for (i = 0; i < (sizeof(streams)/sizeof(*streams)); i++) {
  if ( streams[i].id != -1 ) {
   if ( streams[i].id != execed ) {
    printf("Error: Stream %i is still open: File is corrupted.\n", streams[i].id);
   }
  }
 }

 if ( g_verbose ) {
/*
   printf("multi: %li Bytes in %i (%i empty) data packets (avg: %i byte per packet, %i byte not counting empty packets)\n",
           (long int)datalen, datapkg, datapkg - datapkg_nonempty, datalen/datapkg, datalen/datapkg_nonempty);
*/
  if ( c_data ) {
   printf("multi: %zu Bytes in %zu (%zu empty) data packets "
          "(avg: %zu byte per packet, %zu byte not counting empty packets)\n",
          l_data, c_data, c_data - c_data_nonempty, l_data/c_data, l_data/c_data_nonempty);
  }
#define _HEADER_LEN 10
  printf("multi: %zu (%zu data) Bytes in %zu requests. Overhead is %.2f%%.\n",
         l_all + _HEADER_LEN * reqcount, l_data, reqcount,
         100.f*((float)(l_all + _HEADER_LEN * reqcount)/(float)l_data - 1.f));
 }

 return 0;
}

int main (int argc, char * argv[]) {
 char * k;
 int i;

 for (i = 1; i < argc; i++) {
  k = argv[i];

  if ( !strcmp(k, "-h") || !strcmp(k, "--help") ) {
   usage(argv[0]);
   return 0;
  } if ( !strcmp(k, "-v") || !strcmp(k, "--verbose") ) {
   g_verbose++;
  } else {
   do_file(k);
  }
 }

 if ( argc == 1 ) {
  do_file(NULL);
 }

 return 0;
}

//ll
