//uniraum-play.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2011
 *
 *  This file is part of uniraum,
 *  a cross-platform container library, home and professional use.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  uniraum is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include <uniraum/uniraum.h>

#define CLIENT_NAME "uniraum-play"

#define MAX_STREAMS 8

enum action {
 URP_ADD,
 URP_PROC,
 URP_INIT,
 URP_DESTROY
};

static struct roar_connection g_con;
static int g_is_connected = 0;

struct roar_vio_calls * g_dumpfile = NULL;

static void usage (const char * progname) {
 fprintf(stderr, "Usage: %s [OPTIONS] file [file [file [...]]]\n", progname);
 fprintf(stderr, "\nOptions:\n");
 fprintf(stderr,
                 " -h --help             - Print this help.\n"
                 " -s --server SERVER    - Connect to sound server SERVER.\n"
                 "    --dumpfile FILE    - Dump the data as send to the sound server to FILE.\n"
        );
}

int do_connect(void) {
 const char * server;

 if ( g_is_connected )
  return 0;

 if ( roar_simple_connect(&g_con, NULL, CLIENT_NAME) == 0 )
  return 0;

 server = roar_libroar_get_server();
 ROAR_ERR("Can not connect to %s%s: %s",
                server == NULL ? "default server" : "server: ", server == NULL ? "" : server,
                roar_error2str(roar_error));

 return -1;
}

int do_meta(uniraum_req_t * req) {
 int stream;
 const char * key, * value;
 int mode;
 char prefix[32];

 stream = uniraum_reqget_stream(req);

 if ( uniraum_reqget_meta(req, &key, &value, &mode) == -1 ) {
  ROAR_WARN("File may be corrupted: can not parse meta data.");
  return -1;
 }

 if ( stream == -1 ) {
  snprintf(prefix, sizeof(prefix), "Global: ");
 } else {
  snprintf(prefix, sizeof(prefix), "Stream #%i: ", stream);
 }
 prefix[sizeof(prefix)-1] = 0;

 switch (mode) {
  case ROAR_META_MODE_SET:
    printf("%sMeta data %s set to: %s\n", prefix, key, value);
   break;
  case ROAR_META_MODE_ADD:
    printf("%sMeta data %s added: %s\n", prefix, key, value);
   break;
  case ROAR_META_MODE_CLEAR:
    printf("%sMeta data cleared.\n\n", prefix);
   break;
 }

 return -1;
}

#ifdef TIME_DISPLAY
static const char * time2str(double t, char * buf, size_t len) {
 int h, m;

 if ( t < 0 ) {
//  strncpy(buf, "unknown", len);
//  return buf;
  *buf++ = '-';
  t *= -1;
 }

 h  = t / 3600;
 t -= h * 3600;
 m  = t / 60;
 t -= m * 60;

 snprintf(buf, len, "%.2i:%.2i:%.2i", h, m, (int)t);

 return buf;
}

static void do_show_time(roar_vs_t * vss) {
 double time_total = -1;
 struct roar_stream stream;
 struct roar_audio_info * info = &(stream.info);
 ssize_t pos;
 double time_cur;
 char time_buf[3][10];

 roar_get_stream(roar_vs_connection_obj(vss, NULL), &stream, roar_stream_get_id(roar_vs_stream_obj(vss, NULL)));
 pos = roar_vs_position(vss, ROAR_VS_BACKEND_DEFAULT, NULL);

 if ( pos == -1 ) {
  time_cur = 0;
 } else {
  time_cur = (double)pos/(double)(info->channels*info->rate);
 }

 time2str(time_cur, time_buf[0], sizeof(time_buf[0]));

 if ( time_total > 0 ) {
  time2str(time_total-time_cur, time_buf[1], sizeof(time_buf[1]));
  time2str(time_total, time_buf[2], sizeof(time_buf[2]));

  //Time: 00:02.53 [03:26.20] of 03:28.73  (122.7 kbps)  Output Buffer  43.8%
  fprintf(stderr, "\rTime: %s [%s] of %s            ", time_buf[0], time_buf[1], time_buf[2]);
 } else {
  fprintf(stderr, "\rTime: %s                       ", time_buf[0]);
 }
}
#endif

int do_push_data(roar_vs_t * vss, uniraum_mapping_t * mapping) {
 uniraum_data_t * buf;
 const void * data;
 ssize_t len;
 ssize_t ret;

 while (uniraum_mapping_pending(mapping) > 0) {
  ROAR_DBG("do_push_data(*): There is pending data.");

  buf = uniraum_mapping_dataout(mapping);

  ROAR_DBG("do_push_data(*): buf=%p", buf);

  if ( buf == NULL ) {
   ROAR_WARN("Can not read data buffer from mapper: %s", roar_error2str(uniraum_error(NULL)));
  } else {
   len = uniraum_data_get_len(buf);
   data = uniraum_data_get_data(buf);

   ret = roar_vs_write(vss, data, len, NULL);
   if ( ret != len ) {
    ROAR_WARN("Can not write all data to output: %zi bytes skipped.", (len - ret));
   }

   if ( g_dumpfile != NULL ) {
    ret = roar_vio_write(g_dumpfile, (void*)data, len);
    if ( ret != len ) {
     ROAR_WARN("Can not write all data to dumpfile: %zi bytes skipped.", (len - ret));
    }
   }

#ifdef TIME_DISPLAY
   { static int i = 0;
    if ( ( i % 10 ) == 0 ) {
     do_show_time(vss);
     i = 1;
    } else i++;
   }
#endif
   uniraum_data_unref(buf);
  }
 }

 return 0;
}

int do_stream(enum action action, uniraum_stream_t * stream) {
 static struct {
  uniraum_stream_t  * stream;
  uniraum_mapping_t * mapping;
  roar_vs_t         * vss;
 } streams[MAX_STREAMS];
 struct roar_audio_info vsinfo;
 uniraum_req_t * req;
 uniraum_mapping_t * mapping;
 roar_vs_t * vss;
 uniraum_stream_t * cs;
 size_t i;
 int err;

 // for action=PROC:
 int type;
 int map_req;
 // PROC/NEW_STREAM:
 const struct roar_audio_info * info;
 int dir;

 switch (action) {
  case URP_INIT:
    if ( do_connect() == -1 )
     return -1;

    for (i = 0; i < MAX_STREAMS; i++) {
     streams[i].stream = NULL;
     streams[i].vss    = NULL;
    }
   break;
  case URP_DESTROY:
    for (i = 0; i < MAX_STREAMS; i++) {
     if ( streams[i].stream != NULL )
      uniraum_stream_unref(streams[i].stream);
     if ( streams[i].vss != NULL )
      roar_vs_close(streams[i].vss, ROAR_VS_TRUE, NULL);
     if ( streams[i].mapping != NULL )
      uniraum_mapping_unref(streams[i].mapping);
    }
   break;
  case URP_ADD:
    for (i = 0; i < MAX_STREAMS; i++) {
     if ( streams[i].stream == NULL ) {
      mapping = uniraum_mapping_new(uniraum_stream_get_stream(stream),
                                    UNIRAUM_IN, ROAR_VIO_DFT_RAW,
                                    uniraum_stream_get_auinfo(stream));
      if ( mapping == NULL )
       return -1;

      vss = roar_vs_new_from_con(&g_con, &err);
      if ( vss == NULL ) {
       uniraum_mapping_unref(mapping);
       roar_err_set(err);
       return -1;
      }
      memcpy(&vsinfo, uniraum_stream_get_auinfo(stream), sizeof(vsinfo));
      switch (vsinfo.codec) {
       case ROAR_CODEC_RAUM_VORBIS: vsinfo.codec = ROAR_CODEC_OGG_VORBIS; break;
       case ROAR_CODEC_RAUM_FLAC:   vsinfo.codec = ROAR_CODEC_OGG_FLAC;   break;
      }
      if ( roar_vs_stream(vss, &vsinfo, uniraum_stream_get_dir(stream), &err) == -1 ) {
       roar_vs_close(vss, ROAR_VS_TRUE, NULL);
       uniraum_mapping_unref(mapping);
       roar_err_set(err);
       return -1;
      }

      if ( uniraum_stream_ref(stream) == -1 ) {
       err = roar_error;
       roar_vs_close(vss, ROAR_VS_TRUE, NULL);
       uniraum_mapping_unref(mapping);
       roar_err_set(err);
      }

      streams[i].stream  = stream;
      streams[i].mapping = mapping;
      streams[i].vss     = vss;
      return 0;
     }
    }
    roar_err_set(ROAR_ERROR_MFILE);
    return -1;
   break;
  case URP_PROC:
    for (i = 0; i < MAX_STREAMS; i++) {
     if ( (cs = streams[i].stream) == NULL )
      continue;

     while (cs != NULL && (req = uniraum_stream_reqout(cs)) != NULL) {
      type = uniraum_reqget_type(req);
      map_req = 0;

      switch (type) {
       case ROAR_CMD_SET_META:
         do_meta(req);
        break;
       case ROAR_CMD_NEW_STREAM:
         map_req = 1;
         info = uniraum_stream_get_auinfo(cs);
         dir  = uniraum_stream_get_dir(cs);
         if ( info == NULL || dir == -1 ) {
          ROAR_WARN("File may be corrupted: new stream but can not get basic infos.");
         } else {
          printf("Stream #%i: %uch/%uBit/%uHz, codec=%u(%s), dir=%i(%s) (new stream)\n",
                     uniraum_stream_get_stream(cs),
                     info->channels, info->bits, info->rate,
                     info->codec, roar_codec2str(info->codec),
                     dir, roar_dir2str(dir));
         }
        break;
       case ROAR_CMD_KICK:
         uniraum_mapping_reqin(streams[i].mapping, req);
         do_push_data(streams[i].vss, streams[i].mapping);

         // TODO: this does not work very well for multi-stream RAUM files.
         roar_vs_sync(streams[i].vss, ROAR_VS_WAIT, NULL);

         roar_vs_close(streams[i].vss, ROAR_VS_FALSE, NULL);
         uniraum_mapping_unref(streams[i].mapping);
         uniraum_stream_unref(cs);
         streams[i].stream  = NULL;
         streams[i].mapping = NULL;
         streams[i].vss     = NULL;
         cs                 = NULL;
        break;
       case ROAR_CMD_ADD_DATA:
         map_req = 1;
        break;
      }

      if ( map_req ) {
       if ( uniraum_mapping_reqin(streams[i].mapping, req) == -1 ) {
        ROAR_WARN("Can not map request: %s", roar_error2str(uniraum_error(NULL)));
       }
      }

      uniraum_requnref(req);

      if ( streams[i].vss != NULL && streams[i].mapping != NULL )
       do_push_data(streams[i].vss, streams[i].mapping);
     }
    }
   break;
 }

 return 0;
}

int do_file(const char * filename) {
 uniraum_file_t * inp = NULL;
 uniraum_req_t  * req;
 uniraum_stream_t * streams[MAX_STREAMS];
 uniraum_stream_t * cs;
 int stream, type;
 size_t i;
 int found;

 // for execed streams:
 size_t len;
 ssize_t ret;
 void * data;

 for (i = 0; i < MAX_STREAMS; i++)
  streams[i] = NULL;

 inp = uniraum_opendstr(filename, NULL, O_RDONLY);
 if ( inp == NULL ) {
  ROAR_ERR("Can not open file: %s: %s", filename, roar_error2str(uniraum_error(NULL)));
  return -1;
 }

 do_stream(URP_INIT, NULL);

 while (uniraum_read(inp) >= 0) {
  while ((req = uniraum_reqin(inp)) != NULL) {
   type    = uniraum_reqget_type(req);
   stream  = uniraum_reqget_stream(req);

   if ( stream == -1 ) {
    if ( type == ROAR_CMD_SET_META ) {
     do_meta(req);
    }
   } else {
    cs = NULL;

    if ( type == ROAR_CMD_NEW_STREAM ) {
     if ( (cs = uniraum_stream_new(req)) == NULL ) {
      ROAR_WARN("Can not open new stream #%i: %s", stream, roar_error2str(uniraum_error(NULL)));
     }
     for (i = 0, found = 0; !found && i < MAX_STREAMS; i++) {
      if ( streams[i] == NULL ) {
       streams[i] = cs;
       found = 1;
      }
     }

     if ( i == MAX_STREAMS ) {
      ROAR_WARN("Can not open new stream #%i: %s", stream, roar_error2str(ROAR_ERROR_MFILE));
      uniraum_stream_unref(cs);
      uniraum_requnref(req);
      continue;
     }

     if ( do_stream(URP_ADD, cs) == -1 ) {
      ROAR_WARN("Can not start playback for Stream #%i: %s", stream, roar_error2str(roar_error));
     }
    } else {
     if ( cs == NULL ) {
      for (i = 0; cs == NULL && i < MAX_STREAMS; i++)
       if ( streams[i] != NULL && uniraum_stream_get_stream(streams[i]) == stream )
        cs = streams[i];
     }
     if ( cs == NULL ) {
      ROAR_WARN("File may be corrupted: request for unknown stream #%i. (reqtype=%i)", stream, type);
     } else {
      uniraum_stream_reqin(cs, req);
     }
    }

    if ( type == ROAR_CMD_KICK && cs != NULL ) {
     uniraum_stream_unref(cs);
     for (i = 0; i < MAX_STREAMS; i++)
      if ( streams[i] == cs )
       streams[i] = NULL;
    }
   }

   uniraum_requnref(req);
  }
  do_stream(URP_PROC, NULL);
 }

 if ( (stream = uniraum_get_execed(inp)) != -1 ) {
  cs = NULL;

  for (i = 0; i < MAX_STREAMS; i++) {
   if ( streams[i] != NULL ) {
    if ( uniraum_stream_get_stream(streams[i]) == stream ) {
     cs = streams[i];
    } else {
     uniraum_stream_unref(streams[i]);
     streams[i] = NULL;
    }
   }
  }

  if ( cs == NULL ) {
   ROAR_WARN("File may be corrupted: EXEC to a non existing stream!");
  } else {
   len = 2048;
   req = uniraum_reqdata(stream, 0, NULL, len);
   if ( req == NULL ) {
    ROAR_WARN("Can not create data request: %s", roar_error2str(uniraum_error(NULL)));
   } else {
    data = uniraum_reqget_data(req);
    while ((ret = uniraum_read_execed(inp, data, len)) > 0) {
     uniraum_stream_reqin(cs, req);
     do_stream(URP_PROC, NULL);
    }
    uniraum_requnref(req);
   }
  }
 }

 do_stream(URP_PROC, NULL);
 do_stream(URP_DESTROY, NULL);

 for (i = 0; i < MAX_STREAMS; i++)
  if ( streams[i] != NULL )
   uniraum_stream_unref(streams[i]);

 uniraum_close(inp);

 return 0;
}

int main (int argc, char * argv[]) {
 struct roar_vio_defaults def;
 struct roar_vio_calls dumpfile;
 int i;
 char * k;

 for (i = 1; i < argc; i++) {
  k = argv[i];
  if ( !strcmp(k, "-h") || !strcmp(k, "--help") ) {
   usage(argv[0]);
   return 0;
  } else if ( !strcmp(k, "-s") || !strcmp(k, "--server") ) {
   roar_libroar_set_server(argv[++i]);
  } else if ( !strcmp(k, "--dumpfile") ) {
   if ( roar_vio_dstr_init_defaults(&def, ROAR_VIO_DEF_TYPE_NONE, O_WRONLY|O_CREAT|O_TRUNC, 0644) == -1 ) {
    ROAR_ERR("Can not init DSTR defaults. Bad.");
    return 1;
   }

   if ( roar_vio_open_dstr(&dumpfile, argv[++i], &def, 1) == -1 ) {
    ROAR_ERR("Can not open file: %s: %s", argv[i], roar_error2str(roar_error));
    return 1;
   }

   g_dumpfile = &dumpfile;
  } else {
   do_file(k);
  }
 }

 if ( g_is_connected )
  roar_disconnect(&g_con);

 if ( g_dumpfile != NULL )
  roar_vio_close(g_dumpfile);

 return 0;
}

//ll
